# -*- coding: utf-8 -*-
from math import ceil
import re

from .graver import GRAMMAR_CHARS, GravedDescription, GravedCodeBlock, GravedTable, GravedListing, GravedBlockQuote, \
    GRAVE_TYPE_GLOBAL_REFERENCE, DO_NOT_ADD_SPACE_AFTER, GravedAttributeDescription

INDENT_SIZE_DEFAULT = 4

DO_NOT_PREVIEW_RP = re.compile('this [a-z]+ is a [a-z]+\.?', re.I)

SPACE_CHAR_DEFAULT = ' '
SPACE_CHAR_UNICODE = (b'\xe2\xa0\x80').decode()

DEFAULT_LISTING_HEAD = '-'*(INDENT_SIZE_DEFAULT<<4)

EMBED_DESCRIPTION_MAX_SIZE = 2048

def sizify(words, max_length):
    """
    Builds chunks from the given words, which do not extend the given size.
    
    Parameters
    ----------
    words : `list` of `str`
        The words to build the lines from.
    max_length : `int`
        The maximal length of a generated line.
    
    Returns
    -------
    lines : `list` of `str`
    """
    lines = []
    if words is None:
        return lines
    
    actual_line = []
    line_length = 0
    for word in words:
        word_length = len(word)
        line_length += word_length
        if actual_line:
            line_length += 1
        
        if line_length <= max_length:
            actual_line.append(word)
            continue
        
        if len(actual_line) == 0:
            actual_line.append(word)
            continue
        
        if len(actual_line) == 1:
            line = actual_line.pop(0)
            line_length = 0
        
        else:
            line = ' '.join(actual_line)
            actual_line.clear()
            line_length = word_length
        
        actual_line.append(word)
        lines.append(line)
        continue
    
    if len(actual_line) != 0:
        if len(actual_line) == 1:
            line = actual_line[0]
        else:
            line = ' '.join(actual_line)
        
        lines.append(line)
    
    return lines

def preview_string(graved):
    """
    Creates preview string of the given graved content.
    
    Parameters
    ----------
    graved : `None` or `list` of (`str`, ``Grave``) elements
    
    Returns
    ------
    content : `None` or `str`
    """
    if graved is None:
        return None
    
    words = []
    for element in graved:
        if type(element) is str:
            if words:
                last = words[-1]
                starter = last[0]
                if element[0] in GRAMMAR_CHARS:
                    add_space_before = False
                elif starter == '`':
                    add_space_before = True
                else:
                    add_space_before = True
            else:
                add_space_before = False
            
            if words and words[-1] == ' ':
                if (not add_space_before) and (len(words)>1 and words[-2][-1] in DO_NOT_ADD_SPACE_AFTER) :
                    del words[-1]
        
        else:
            element = element.content
        
        words.append(element)
        words.append(' ')
    
    if words:
        del words[-1]
        content = ''.join(words)
    else:
        content = None
    
    return content

def graved_to_single_graved_words(graved):
    """
    Translates the given graved content to words.
    
    Parameters
    ----------
    graved : `None` or `list` of (`str`, ``Grave``) elements
    
    Returns
    -------
    words : `None` or `list` of `str`
    """
    if graved is None:
        return None
    
    words = []
    for element in graved:
        if type(element) is str:
            if words:
                last = words[-1]
                starter = last[0]
                if element[0] in GRAMMAR_CHARS:
                    add_space_before = False
                elif starter == '`':
                    add_space_before = True
                else:
                    add_space_before = True
            else:
                add_space_before = False
            
            local_words = element.split(' ')
            if (not add_space_before) and words:
                words[-1] = words[-1]+local_words.pop(0)
            
            words.extend(local_words)
        
        else:
            local_word = f'`{element.content}`'
            words.append(local_word)
    
    return words

def graved_to_source_words(graved):
    """
    Translates the given graved content back to source like words.
    
    Parameters
    ----------
    graved : `None` or `list` of (`str`, ``Grave``) elements
    
    Returns
    -------
    words : `None` or `list` of `str`
    """
    if graved is None:
        return None
    
    words = []
    for element in graved:
        if type(element) is str:
            if words:
                last = words[-1]
                starter = last[0]
                if element[0] in GRAMMAR_CHARS:
                    add_space_before = False
                elif starter == '`':
                    add_space_before = True
                else:
                    add_space_before = True
            else:
                add_space_before = False
            
            local_words = element.split(' ')
            if (not add_space_before) and words:
                words[-1] = words[-1]+local_words.pop(0)
            
            words.extend(local_words)
            
        else:
            if element.type == GRAVE_TYPE_GLOBAL_REFERENCE:
                graver = '``'
            else:
                graver = '`'
            
            local_word = f'{graver}{element.content}{graver}'
            words.append(local_word)
    
    return words

def graved_to_escaped_words(graved):
    """
    Translates the given graved content to words where underscores are escaped..
    
    Parameters
    ----------
    graved : `None` or `list` of (`str`, ``Grave``) elements
    
    Returns
    -------
    words : `None` or `list` of `str`
    """
    if graved is None:
        return None
    
    words = []
    for element in graved:
        if type(element) is str:
            if words:
                last = words[-1]
                starter = last[0]
                if element[0] in GRAMMAR_CHARS:
                    add_space_before = False
                elif starter == '`':
                    add_space_before = True
                else:
                    add_space_before = True
            else:
                add_space_before = False
            
            local_words = element.replace('_', '\_').split(' ')
            if (not add_space_before) and words:
                words[-1] = words[-1]+local_words.pop(0)
            
            words.extend(local_words)
        
        else:
            local_word = f'`{element.content}`'
            words.append(local_word)
    
    return words

class BuilderContext:
    """
    Builder context for converters.
    
    Parameters
    ----------
    indent : `int`
        A pregenerated indention.
    indent_size : `int`
        The size of an indent.
    max_line_length : `int`
        The maximal length of a generated line.
    space_char : `str`
        The character what is used when building tables and indents.
    word_converter : `function`
        Converter to transform graved content to words.
    """
    __slots__ = ('indent', 'indent_size', 'max_line_length', 'space_char', 'word_converter')
    def __init__(self, indent_size, space_char, max_line_length, word_converter):
        self.indent_size = indent_size
        self.space_char = space_char
        self.word_converter = word_converter
        self.max_line_length = max_line_length
        self.indent = indent_size*space_char
    
    def __repr__(self):
        """Returns the builder context's representation."""
        return (f'{self.__class__.__name__}(indent_size={self.indent_size!r}, max_line_length='
            f'{self.max_line_length!r}, space_char={self.space_char!r}, word_converter={self.word_converter!r})')

EMBED_SIZED_BUILDER_CONTEXT = BuilderContext(INDENT_SIZE_DEFAULT, SPACE_CHAR_UNICODE, 79, graved_to_escaped_words)
TEXT_BUILDER_CONTEXT = BuilderContext(INDENT_SIZE_DEFAULT, SPACE_CHAR_DEFAULT, 120, graved_to_single_graved_words)
TEXT_SOURCE_BUILDER_CONTEXT = BuilderContext(INDENT_SIZE_DEFAULT, SPACE_CHAR_DEFAULT, 120, graved_to_source_words)

class TableLine:
    """
    Represents a line inside of a table.
    
    Attributes
    ----------
    lines : `list` of `str`
        The lines, to what the respective table row is broken down.
    """
    def __init__(self, indent_level, line, sizes, builder_context):
        """
        Creates a new table line instance.
        
        Parameters
        ----------
        indent_level : `int`
            How much the respective table is indented.
        line : `list` of (`None`, `str`, `list` of `str`)
            The generated line(s) of the represented line of the table.
        sizes : `list` of `int`
            The internal size for each column.
        builder_context : ``BuilderContext``
            Context to define some building details.
        """
        longest = 1
        for line_part in line:
            if type(line_part) is str:
                continue
            
            length = len(line_part)
            if length > longest:
                longest = length
        
        lines = []
        
        for index in range(longest):
            # Add indent
            result_line = []
            for _ in range(indent_level):
                result_line.append(builder_context.indent)
            
            # Add starter
            result_line.append('|')
            
            for column_index in range(len(line)):
                result_line.append(builder_context.space_char)
                part = line[column_index]
                if part is None:
                    result_line.append(builder_context.space_char*sizes[column_index])
                
                if type(part) is str:
                    if index:
                        result_line.append(builder_context.space_char*sizes[column_index])
                    else:
                        result_line.append(part)
                        result_line.append(builder_context.space_char*(sizes[column_index]-len(part)))
                        continue
                
                if len(part) <= index:
                    result_line.append(builder_context.space_char*sizes[column_index])
                else:
                    part = part[index]
                    result_line.append(part)
                    result_line.append(builder_context.space_char*(sizes[column_index]-len(part)))
                
                result_line.append(builder_context.space_char)
                result_line.append('|')
            
            result_line.append('\n')
            generated_line = ''.join(result_line)
            lines.append(generated_line)
        
        self.lines = lines
    
    @property
    def line_count(self):
        """
        Returns how much lines the given row is.
        
        Returns
        -------
        length : `int`
        """
        return len(self.lines)
    
    @property
    def character_count(self):
        """
        Returns how much character the given row is from.
        
        Returns
        -------
        length : `int`
        """
        length = 0
        for line in self.lines:
            length += len(line)
        
        return length

class TableConverter:
    """
    Converter class for tables, when building text.
    
    Attributes
    ----------
    head_under_line : `str`
        Separator line after the first line of the table.
    lines : `list` of `str`
        The lines of the table.
    separator_line : `str`
        Separator line between all lines of the table (expect the first two).
    """
    __slots__ = ('head_under_line', 'lines', 'separator_line')
    def __new__(cls, table, indent_level, optimal_fit, builder_context):
        """
        Creates a new table to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        table : ``GravedTable``
            The source table.
        indent_level : `int`
            The number of how far is the table indented.
        optimal_fit : `int`
            The preferred maximal line length of the table.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``TableConverter``
        """
        x, y = table.size
        lines = []
        for line in table:
            new_line = [builder_context.word_converter(part) for part in line]
            lines.append(new_line)
        
        # calculate part lengths for each column
        longest_lengths = [0 for _ in range(x)]
        
        for line in lines:
            for index in range(x):
                part = line[index]
                if part is None:
                    continue
                
                length = len(part) -1
                for word in part:
                    length +=len(word)
                
                if longest_lengths[index] < length:
                    longest_lengths[index] = length
        
        fillable_chars = optimal_fit - (x+1) - ((x-1)<<1)
        
        # do we fit?
        if sum(longest_lengths) <= fillable_chars:
            sizes = longest_lengths
            for index in range(len(lines)):
                line = lines[index]
                if line is None:
                    continue
                
                lines[index] = [[' '.join(parts)] for parts in line]
        else:
            # calculate minimal lengths, aka longest word's length for each column
            shortest_lengths = [0 for _ in range(x)]
            for line in lines:
                for index in range(x):
                    part = line[index]
                    if part is None:
                        continue
                    
                    length = 0
                    
                    for word in part:
                        word_length = len(word)
                        if length < word_length:
                            length = word_length
                    
                    if shortest_lengths[index] < length:
                        shortest_lengths[index] = length
            
            # calculate average lengths
            average_lengths = [0 for _ in range(x)]
            for line in lines:
                for index in range(x):
                    part = line[index]
                    if part is None:
                        continue
                    
                    length = len(line) -1
                    for word in part:
                        length +=len(word)
                    
                    longest_lengths[index] += length
            
            for index in range(x):
                length =  average_lengths[index] / y
                length = ceil(length)
                average_lengths[index] = length
            
            size_factors = [
                (((average_lengths[index]<<1) + longest_lengths[index])*shortest_lengths[index])**.5 \
                    for index in range(x)]
            
            factorizable_chars = fillable_chars - sum(shortest_lengths)
            
            factor = factorizable_chars / fillable_chars
            
            sizes = [ceil(size_factors[index]*factor + shortest_lengths[index]) for index in range(x)]
            
            for line in lines:
                for index in range(x):
                    line[index] = sizify(line[index], sizes[index])
            
            sizes = [0 for _ in range(x)]
            
            for line in lines:
                for index in range(x):
                    line_parts = line[index]
                    for part in line_parts:
                        length = len(part)
                        
                        if sizes[index] < length:
                            sizes[index] = length
        
        generated_lines = [TableLine(indent_level, line, sizes, builder_context) for line in lines]
        
        result_line = []
        for _ in range(indent_level):
            result_line.append(builder_context.indent)
        
        for length in sizes:
            result_line.append('+')
            result_line.append('='*(length+2))
        
        result_line.append('+\n')
        
        head_under_line = ''.join(result_line)
        result_line.clear()
        for _ in range(indent_level):
            result_line.append(builder_context.indent)
        
        for length in sizes:
            result_line.append('+')
            result_line.append('-'*(length+2))
        
        result_line.append('+\n')
        
        separator_line = ''.join(result_line)
        
        self = object.__new__(cls)
        self.head_under_line = head_under_line
        self.lines = generated_lines
        self.separator_line = separator_line
        yield self
    
    @property
    def character_count(self):
        """
        Returns the table's characters' total length.
        
        Returns
        -------
        length : `int`
        """
        lines = self.lines
        length = (len(lines)+1)*len(self.separator_line)
        for line in lines:
            length += line.character_count
        
        return length
    
    @property
    def row_count(self):
        """
        Returns the table's row count.
        
        Returns
        -------
        length : `int`
        """
        return len(self.lines)
    
    @property
    def line_count(self):
        """
        Returns the table's total line count.
        
        Returns
        -------
        length : `int`
        """
        lines = self.lines
        length = len(lines)+1
        for line in lines:
            length += line.line_count
        
        return length
    
    def _do_break(self, number_of_rows):
        """
        Does break after the given amount of rows.
        
        Parameters
        ----------
        number_of_rows : `int`
            The number of rows after the table should be broken.
        
        Returns
        -------
        table_1 : ``TableConverter``
        table_2 : ``TableConverter``
        """
        number_of_rows +=1
        lines = self.lines
        head = lines[0]
        
        table_1 = object.__new__(type(self))
        table_1.head_under_line = self.head_under_line
        table_1.lines = [head, *lines[1:number_of_rows]]
        table_1.separator_line = self.separator_line
        
        table_2 = object.__new__(type(self))
        table_2.head_under_line = self.head_under_line
        table_2.lines = [head, *lines[number_of_rows:]]
        table_2.separator_line = self.separator_line
        
        return table_1, table_2
    
    def _test_break(self):
        """
        Returns whether this table should be rendered alone as a broken part of a bigger table.
        
        Returns
        -------
        passed : `bool`
        """
        if self.row_count < 2:
            return False
        
        if self.line_count < 7:
            return False
        
        return True
    
    def do_break(self, number_of_chars):
        """
        Breaks the table to parts an returns the most optimal case for the given max chars.
        
        Parameters
        ----------
        number_of_chars : `int`
            The maximal amount of chars, what the first table should contain.
        
        Returns
        -------
        best_fit : `None` or `tuple` (``TableConverter``, ``TableConverter``)
            The best fitting 2 table shard, if there is any optimal case. If there is non, returns `None` instead.
        """
        best_fit = None
        for number_of_rows in range(1, len(self.lines)-2):
            tables = self._do_break(number_of_rows)
            table_1, table_2 = tables
            if table_1.character_count > number_of_chars:
                break
            
            if not table_1._test_break():
                continue
            
            if not table_2._test_break():
                continue
            
            best_fit = tables
            continue
        
        return best_fit
    
    def render_to(self, to_extend):
        """
        Renders the table's lines to the given `list`.
        
        Parameters
        ----------
        to_extend : `list` of `str`
            A list to what the table should yield it's lines.
        """
        separator_line = self.separator_line
        to_extend.append(separator_line)
        lines = self.lines
        to_extend.extend(lines[0].lines)
        if len(lines) == 1:
            to_extend.append(separator_line)
            return
        
        to_extend.append(self.head_under_line)
        for line in lines[1:]:
            to_extend.extend(line.lines)
            to_extend.append(separator_line)
    
    def __repr__(self):
        """Returns the table's representation."""
        to_extend = [self.__class__.__name__, ':\n']
        self.render_to(to_extend)
        to_extend.append('<<END>>')
        return ''.join(to_extend)

class CodeBlockConverter:
    """
    Converter class for code blocks, when building text.
    
    Attributes
    ----------
    ender : `str`
        The last line of the code block.
    lines : `list` of `str`
        The lines of the code block.
    starter : `str`
        The first line of the code block.
    """
    __slots__ = ('ender', 'lines', 'starter')
    def __new__(cls, code_block, indent_level, optimal_fit, builder_context):
        """
        Creates a new code block to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        code_block : ``GravedCodeBlock``
            The source code block.
        indent_level : `int`
            The number of how far is the code block indented.
        optimal_fit : `int`
            The preferred maximal line length of the code block.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``CodeBlockConverter``
        """
        indent = builder_context.indent*indent_level
        self = object.__new__(cls)
        self.lines = [f'{indent}{line}\n' for line in code_block.lines]
        ender = indent+'```'
        language = code_block.language
        if language is None:
            starter = ender
        else:
            starter = ender+language
        
        self.starter = starter
        self.ender = ender
        yield self
    
    @property
    def character_count(self):
        """
        Returns the code block's characters' total length.
        
        Returns
        -------
        length : `int`
        """
        lines = self.lines
        length = len(self.starter)+len(self.ender)
        for line in lines:
            length += len(line)
        
        return length
    
    @property
    def interactive_row_count(self):
        """
        Returns the code block representative interactive shell row count.
        
        Returns
        -------
        length : `int`
        """
        length = 0
        for line in self.lines:
            if line.startswith('>>>'):
                length += 1
        
        return length
    
    @property
    def code_row_count(self):
        """
        Returns how much line break is in the text.
        
        Returns
        -------
        length : `int`
        """
        last_line_void = False
        length = 0
        for line in self.lines:
            if line:
                last_line_void = False
                continue
            
            if last_line_void:
                continue
            
            last_line_void = True
            length +=1
        
        return length
    
    @property
    def line_count(self):
        """
        Returns the code block's total line count.
        
        Returns
        -------
        length : `int`
        """
        return 2+len(self.lines)
    
    @classmethod
    def _create_remove_empty_lines(cls, lines, starter, ender):
        """
        Creates a new code block from the given lines and indent level.
        
        Removes the empty lines from the start and end.
        
        Parameters
        ----------
        lines : `list` of `str`
            The code block's lines.
        starter : `str`
            The first line of the code block.
        ender : `str`
            The last line of the code block
        
        Returns
        -------
        self : ``CodeBlockConverter``
        """
        while True:
            if not lines:
                break
            
            if lines[-1]:
                while True:
                    if not lines:
                        break
                    
                    if lines[0]:
                        break
                    
                    del lines[0]
                    continue
                break
            
            del lines[-1]
            continue
        
        self = object.__new__(cls)
        self.lines = lines
        self.starter = starter
        self.ender = ender
        return self
    
    def _do_interactive_break(self, number_of_rows):
        """
        Does a break at the given interactive row input-output.
        
        Parameters
        ----------
        number_of_rows : `int`
            The last interactive row, what should be included in the first code block.
        
        Returns
        -------
        code_block_1 : ``CodeBlockConverter``
        code_block_2 : ``CodeBlockConverter``
        """
        rows_found = -1
        lines = self.lines
        
        index = 0
        limit = len(lines)
        while True:
            if index == limit:
                code_block_1 = self
                self._create_remove_empty_lines([], self.starter, self.ender)
                break
            
            line = lines[index]
            if not line.startswith('>>>'):
                index +=1
                continue
            
            rows_found +=1
            if rows_found != number_of_rows:
                index +=1
                continue
            
            code_block_1 = self._create_remove_empty_lines(lines[:index], self.starter, self.ender)
            code_block_2 = self._create_remove_empty_lines(lines[index:], self.starter, self.ender)
            break
        
        return code_block_1, code_block_2
    
    def _test_break(self):
        """
        Returns whether this code block should be rendered alone as a broken part of a bigger one.
        
        Returns
        -------
        passed : `bool`
        """
        # At least 4 lines.
        if len(self.lines < 3):
            return False
        
        return True
        
    def _do_code_break(self, number_of_rows):
        """
        Does a break at the given interactive row input-output.
        
        Parameters
        ----------
        number_of_rows : `int`
            The last interactive row, what should be included in the first code block.
        
        Returns
        -------
        code_block_1 : ``CodeBlockConverter``
        code_block_2 : ``CodeBlockConverter``
        """
        rows_found = -1
        lines = self.lines
        last_line_void = False
        
        index = 0
        limit = len(lines)
        while True:
            if index == limit:
                code_block_1 = self
                self._create_remove_empty_lines([], self.starter, self.ender)
                break
            
            line = lines[index]
            if line:
                last_line_void = False
                index +=1
                continue
            
            if last_line_void:
                index +=1
                continue
            
            rows_found +=1
            if rows_found != number_of_rows:
                index +=1
                continue
            
            code_block_1 = self._create_remove_empty_lines(lines[:index], self.starter, self.ender)
            code_block_2 = self._create_remove_empty_lines(lines[index:], self.starter, self.ender)
            break
        
        return code_block_1, code_block_2
    
    def _do_line_break(self, number_of_rows):
        """
        Does a break at the given line.
        
        Parameters
        ----------
        number_of_rows : `int`
            The first line's index, what should be not included at the first code block.
        
        Returns
        -------
        code_block_1 : ``CodeBlockConverter``
        code_block_2 : ``CodeBlockConverter``
        """
        code_block_1 = self._create_remove_empty_lines(self.lines[:number_of_rows], self.starter, self.ender)
        code_block_2 = self._create_remove_empty_lines(self.lines[number_of_rows:], self.starter, self.ender)
        return code_block_1, code_block_2
    
    def do_break(self, number_of_chars):
        """
        Breaks the code block to parts an returns the most optimal case for the given max chars.
        
        Parameters
        ----------
        number_of_chars : `int`
            The maximal amount of chars, what the first code block should contain.
        
        Returns
        -------
        best_fit : `None` or `tuple` (``CodeBlockConverter``, ``CodeBlockConverter``)
            The best fitting 2 code block shard, if there is any optimal case. If there is non, returns `None` instead.
        """
        
        if len(self.lines) < 6:
            return None
        
        length = self.interactive_row_count
        if length > 2:
            best_fit = None
            
            for x in range(1, length-1):
                code_blocks = self._do_interactive_break(x)
                code_block_1, code_block_2 = code_blocks
                
                if code_block_1.character_count > number_of_chars:
                    break
                
                if not code_block_1._test_break():
                    continue
                
                if not code_block_2._test_break():
                    continue
                
                best_fit = code_blocks
                continue
            
            if (best_fit is not None):
                return best_fit
        
        else:
            length = self.code_row_count
            if length > 2:
                best_fit = None
                
                for x in range(1, length-1):
                    code_blocks = self._do_code_break(x)
                    code_block_1, code_block_2 = code_blocks
                    
                    if code_block_1.character_count > number_of_chars:
                        break
                    
                    if not code_block_1._test_break():
                        continue
                    
                    if not code_block_2._test_break():
                        continue
                    
                    best_fit = code_blocks
                    continue
            
                if (best_fit is not None):
                    return best_fit
        
        if number_of_chars < 600:
            return None
        
        best_fit = None
        
        for x in range(3, len(self.lines)-3):
            code_blocks = self._do_line_break(x)
            
            if code_blocks[0].character_count > number_of_chars:
                break
            
            best_fit = code_blocks
            continue
            
        return best_fit
    
    def render_to(self, to_extend):
        """
        Renders the code block's lines to the given `list`.
        
        Parameters
        ----------
        to_extend : `list` of `str`
            A list to what the code block should yield it's lines.
        """
        to_extend.append(self.starter)
        to_extend.extend(self.lines)
        to_extend.append(self.ender)
    
    def __repr__(self):
        """Returns the code block's representation."""
        to_extend = [self.__class__.__name__, ':\n']
        self.render_to(to_extend)
        to_extend.append('<<END>>')
        return ''.join(to_extend)


class DescriptionConverter:
    """
    Converter class for descriptions when building text.
    
    Attributes
    ----------
    lines : `list` of `str`
        The lines of the description.
    indent : `int`
        The indent level of the description.
    """
    __slots__ = ('indent', 'lines', )
    def __new__(cls, description, indent_level, optimal_fit, builder_context):
        """
        Creates a new description to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        description : ``GravedDescription``
            The source description.
        indent_level : `int`
            The number of how far is the description indented.
        optimal_fit : `int`
            The preferred maximal line length of the description.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``DescriptionConverter``
        """
        words = builder_context.word_converter(description.content)
        lines  = sizify(words, optimal_fit)
        
        indention = builder_context.indent*indent_level
        lines = [f'{indention}{line}\n' for line in lines]
        
        self = object.__new__(cls)
        self.lines = lines
        self.indent = indent_level
        yield self
    
    @property
    def character_count(self):
        """
        Returns the description's characters' total length.
        
        Returns
        -------
        length : `int`
        """
        length = 0
        for line in self.lines:
            length +=len(line)
        
        return length
    
    @property
    def line_count(self):
        """
        Returns the description's total line count.
        
        Returns
        -------
        length : `int`
        """
        return len(self.lines)
    
    def _do_break(self, number_of_rows):
        """
        Breaks the description after the given amount of rows.
        
        Parameters
        ----------
        number_of_rows : `int`
            The number of rows before the description should be broken.
        
        Returns
        -------
        description_1 : ``DescriptionConverter``
        description_2 : ``DescriptionConverter``
        """
        description_1 = object.__new__(type(self))
        description_1.lines = self.lines[:number_of_rows]
        description_2 = object.__new__(type(self))
        description_2.lines = self.lines[number_of_rows:]
    
    def _test_break(self):
        """
        Returns whether this description should be rendered alone as a broken part of a bigger one.
        
        Returns
        -------
        passed : `bool`
        """
        if len(self.lines) < 4:
            return False
        
        if len(self.character_count) < 300:
            return False
        
        return True
    
    def do_break(self, number_of_chars):
        """
        Breaks the description to pats and returns the most optimal case for the given max chars.
        
        Parameters
        ----------
        number_of_chars : `int`
            The maximal amount of chars, what the first description should contain.

        Returns
        -------
        best_fit : `None` or `tuple` (``DescriptionConverter``, ``DescriptionConverter``)
            The best fitting 2 description shard, if there is any optimal case. If there is non, returns `None` instead.
        """
        lines = self.lines
        best_fit = None
        for x in range(4, len(lines)-4):
            descriptions = self._do_break(x)
            description_1, description_2 = descriptions
            
            if description_1.character_count > number_of_chars:
                break
            
            if not description_1._test_break():
                continue
            
            if not description_2._test_break():
                continue
            
            best_fit = descriptions
            continue
        
        return best_fit
    
    def render_to(self, to_extend):
        """
        Renders the description's lines to the given `list`.
        
        Parameters
        ----------
        to_extend : `list` of `str`
            A list to what the description should yield it's lines.
        """
        to_extend.extend(self.lines)
    
    def __repr__(self):
        """Returns the description's representation."""
        to_extend = [self.__class__.__name__, ':\n']
        self.render_to(to_extend)
        to_extend.append('<<END>>')
        return ''.join(to_extend)


class BlockQuoteConverter:
    """
    Converter class for descriptions when building text.
    
    Attributes
    ----------
    empty_line : `str`
        Empty line inside of the block quote separating 2 paragraphs.
    indent : `int`
        The indent level of the block quote.
    lines_by_paragraph : `list` of (`list` of `str`)
        The lines of the block quote.
    """
    __slots__ = ('empty_line', 'indent', 'lines_by_paragraph', )
    def __new__(cls, block_quote, indent_level, optimal_fit, builder_context):
        """
        Creates a new block quote to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        block_quote : ``GravedBlockQuote``
            The source block quote.
        indent_level : `int`
            The number of how far is the block quote indented.
        optimal_fit : `int`
            The preferred maximal line length of the block quote.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``BlockQuoteConverter``
        """
        lines_by_paragraph = []
        indention = builder_context.indent*indent_level
        
        for description in block_quote.descriptions:
            description_words = builder_context.word_converter(description)
            description_lines = sizify(description_words, optimal_fit)
            lines = [f'{indention}> {line}\n' for line in description_lines]
            lines_by_paragraph.append(lines)
        
        self = object.__new__(cls)
        self.lines_by_paragraph = lines_by_paragraph
        self.indent = indent_level
        self.empty_line = f'{indention}>\n'
        yield self
    
    @property
    def character_count(self):
        """
        Returns the block quote's characters' total length.
        
        Returns
        -------
        length : `int`
        """
        lines_by_paragraphs = self.lines_by_paragraph
        length = (len(lines_by_paragraphs)-1)*len(self.empty_line)
        
        for lines in lines_by_paragraphs:
            for line in lines:
                length += len(line)
        
        return length
    
    @property
    def line_count(self):
        """
        Returns the block quote's total line count.
        
        Returns
        -------
        length : `int`
        """
        lines_by_paragraphs = self.lines_by_paragraph
        length = len(lines_by_paragraphs)-1
        for lines in lines_by_paragraphs:
            length += len(lines)
        
        return length
    
    def _do_break(self, number_of_rows):
        """
        Breaks the block quote after the given amount of rows.
        
        Parameters
        ----------
        number_of_rows : `int`
            The number of rows before the block quote should be broken.
        
        Returns
        -------
        block_quote_1 : ``BlockQuoteConverter``
        block_quote_2 : ``BlockQuoteConverter``
        """
        block_quote_1_lines_by_paragraph = []
        block_quote_2_lines_by_paragraph = []
        
        break_found = False
        for lines in self.lines_by_paragraph:
            if break_found:
                block_quote_2_lines_by_paragraph.append(lines)
                continue
            
            lines_count = len(lines)
            if lines_count > number_of_rows:
                block_quote_1_lines_by_paragraph.append(lines)
                number_of_rows -= lines_count
                continue
            
            if lines_count == number_of_rows:
                block_quote_1_lines_by_paragraph.append(lines)
                break_found = True
                continue
            
            block_quote_1_lines_by_paragraph.append(lines[:number_of_rows])
            block_quote_2_lines_by_paragraph.append(lines[number_of_rows:])
            break_found = True
            continue
        
        
        block_quote_1 = object.__new__(type(self))
        block_quote_1.lines_by_paragraph = block_quote_1_lines_by_paragraph
        block_quote_1.indent = self.indent
        block_quote_1.empty_line = self.empty_line
        block_quote_2 = object.__new__(type(self))
        block_quote_2.lines_by_paragraph = block_quote_2_lines_by_paragraph
        block_quote_1.indent = self.indent
        block_quote_2.empty_line = self.empty_line
        return block_quote_1, block_quote_2
    
    def _test_break(self):
        """
        Returns whether this block quote should be rendered alone as a broken part of a bigger one.
        
        Returns
        -------
        passed : `bool`
        """
        lines_by_paragraph = self.lines_by_paragraph
        if lines_by_paragraph:
            return True
        
        if len(self.line_count) < 4:
            return False
        
        if len(self.character_count) < 300:
            return False
        
        return True
    
    def do_break(self, number_of_chars):
        """
        Breaks the block quote to pats and returns the most optimal case for the given max chars.
        
        Parameters
        ----------
        number_of_chars : `int`
            The maximal amount of chars, what the first block quote should contain.
        
        Returns
        -------
        best_fit : `None` or `tuple` (``BlockQuoteConverter``, ``BlockQuoteConverter``)
            The best fitting 2 block quote shards, if there is any optimal case. If there is non, returns `None`
            instead.
        """
        best_fit = None
        
        line_count_without_separators = 0
        lines_by_paragraph = self.lines_by_paragraph
        for lines in lines_by_paragraph:
            line_count_without_separators += len(lines)
        
        lines = lines_by_paragraph[0]
        start_index = len(lines)
        if start_index > 4:
            start_index = start_index
        
        end_index = lines_by_paragraph[-1]
        if end_index > 4:
            end_index = 4
        
        for x in range(start_index, line_count_without_separators-end_index):
            block_quotes = self._do_break(x)
            block_quote_1, block_quote_2 = block_quotes
            
            if block_quote_1.character_count > number_of_chars:
                break
            
            if not block_quote_1._test_break():
                continue
            
            if not block_quote_2._test_break():
                continue
            
            best_fit = block_quotes
            continue
        
        return best_fit
    
    def render_to(self, to_extend):
        """
        Renders the block quote's lines to the given `list`.
        
        Parameters
        ----------
        to_extend : `list` of `str`
            A list to what the description should yield it's lines.
        """
        lines_by_paragraph = self.lines_by_paragraph
        
        index = 0
        limit = len(lines_by_paragraph)
        
        while True:
            lines = lines_by_paragraph[index]
            index += 1
            
            to_extend.extend(lines)
            
            if index == limit:
                break
            
            to_extend.append(self.lines)
    
    def __repr__(self):
        """Returns the block quote's representation."""
        to_extend = [self.__class__.__name__, ':\n']
        self.render_to(to_extend)
        to_extend.append('<<END>>')
        return ''.join(to_extend)


class AttributeDescriptionConverter(DescriptionConverter):
    """
    Converter class for attribute descriptions, when when building text.
    
    Attributes
    ----------
    lines : `list` of `str`
        The lines of the listing element's head.
    indent : `int`
        The indent level of the description.
    """
    def __new__(cls, attribute_description, indent_level, optimal_fit, builder_context):
        """
        Creates a new description to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        attribute_description : ``GravedAttributeDescription``
            The source description.
        indent_level : `int`
            The number of how far is the description indented.
        optimal_fit : `int`
            The preferred maximal line length of the description.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``AttributeDescriptionConverter``
        """
        separator = attribute_description.separator
        
        if separator == '(':
            spacing = ''
        else:
            spacing = ' '
        
        words = [f'{attribute_description.name}{spacing}{separator}']
        
        words.extend(builder_context.word_converter(attribute_description.content))
        lines  = sizify(words, optimal_fit)
        
        indention = builder_context.indent*indent_level
        lines = [f'{indention}{line}\n' for line in lines]
        
        self = object.__new__(cls)
        self.lines = lines
        self.indent = indent_level
        yield self

def listing_converter(listing, indent_level, optimal_fit, builder_context):
    """
    Deserializes the given listing to converters.
    
    This function is a generator.
    
    Parameters
    ----------
    listing : ``GravedListing``
        The source listing.
    indent_level : `int`
        The number of how far is the listing indented.
    optimal_fit : `int`
        The preferred maximal line length of the listing.
    builder_context : ``BuilderContext``
        Context to define some building details.
    
    Yields
    -----------
    element : `Any`
    """
    for listing_element in listing.elements:
        yield from ListingHeadConverter(listing_element.head, indent_level, optimal_fit, builder_context)
        content = listing_element.content
        if content is None:
            continue
        
        yield from sub_section_converter(content, indent_level+1, optimal_fit-builder_context.indent_size, builder_context)


class ListingHeadConverter(DescriptionConverter):
    """
    Converter class for listing element heads when building text.
    
    Attributes
    ----------
    lines : `list` of `str`
        The lines of the listing element's head.
    indent : `int`
        The indent level of the description.
    """
    def __new__(cls, head, indent_level, optimal_fit, builder_context):
        """
        Creates a new listing head to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        head : `None` or `list` of (`str`, ``Grave``)
            The head of a listing.
        indent_level : `int`
            The number of how far is the listing element indented.
        optimal_fit : `int`
            The preferred maximal line length.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Yields
        ------
        self : ``ListingHeadConverter``
        """
        new_lines = []
        if head is None:
            lines = [DEFAULT_LISTING_HEAD]
        else:
            words = builder_context.word_converter(head)
            lines = sizify(words, optimal_fit-builder_context.indent_size)
        
        new_lines.append(f'{(builder_context.indent*indent_level)}- {lines[0]}\n')
        
        if len(lines)>1:
            indention = builder_context.indent*(indent_level+1)
            for line in lines[1:]:
                new_lines.append(f'{indention}{line}\n')
        
        self = object.__new__(cls)
        self.lines = new_lines
        self.indent = indent_level
        yield self


def sub_section_converter(sub_section, indent_level, optimal_fit, builder_context):
    """
    Deserializes the given sub-section to converters.
    
    This method is a generator.
    
    Parameters
    ----------
    sub_section : `list` of `Any`
        The source sub-section..
    indent_level : `int`
        The number of how far is the sub-section indented.
    optimal_fit : `int`
        The preferred maximal line length of the sub-section.
    builder_context : ``BuilderContext``
        Context to define some building details.
    
    Yields
    -----------
    element : `Any`
    """
    for element in sub_section:
        # If the converter is a sub section converter, indent it
        converter = CONVERTER_TABLE[type(element)]
        if converter is sub_section_converter:
            local_indent_level = indent_level+1
            local_optimal_fit = optimal_fit-builder_context.indent_size
        else:
            local_indent_level = indent_level
            local_optimal_fit = optimal_fit
        
        yield from converter(element, local_indent_level, local_optimal_fit, builder_context)


CONVERTER_TABLE = {
    list : sub_section_converter,
    GravedListing : listing_converter,
    GravedDescription : DescriptionConverter,
    GravedAttributeDescription : AttributeDescriptionConverter,
    GravedTable : TableConverter,
    GravedCodeBlock : CodeBlockConverter,
    GravedBlockQuote : BlockQuoteConverter,
        }


class SectionTitleConverter(DescriptionConverter):
    """
    Converter class for section titles when building a text.
    
    Attributes
    ----------
    lines : `list` of `str`
        The lines of the section title.
    """
    def __new__(cls, title):
        """
        Creates a new section title to text converter.
        
        This method is a generator.
        
        Parameters
        ----------
        title : `None` or `str`
            The respective section's title.
        
        Yields
        ------
        self : `None` or ``DescriptionConverter``
        """
        if title is None:
            return
        
        self = object.__new__(cls)
        self.lines = [title+'\n', '-'*len(title)+'\n']
        self.indent = -2
        yield self

def should_insert_linebreak(part_1, part_2):
    """
    Returns whether there should be a linebreak inserted between the two given section parts.
    
    Parameters
    ----------
    part_1 : `Any`
        The last converter.
    part_2 : `Any`
        The next converter.
    
    Returns
    -------
    should_insert : `bool`
    """
    if not isinstance(part_1, DescriptionConverter):
        return True
    
    if isinstance(part_1, SectionTitleConverter):
        return False
    
    if not isinstance(part_2, DescriptionConverter):
        return False
    
    if part_1.indent+1 == part_2.indent:
        return False
    
    return True

def should_accept_section_break(section_1, section_2):
    """
    Returns whether the two section is correctly broken into two parts.
    
    Note that, `section_1._test_break()` and `section_2._test_break()` should be already handled before this function is
    called.
    
    Parameters
    ----------
    section_1 : ``SectionConverter``
        The first converter.
    section_2 : ``SectionConverter``
        The second converter.
    
    Returns
    -------
    should_accept : `bool`
    """
    section_1_last = section_1.parts[-1]
    section_2_first = section_2.parts[0]
    
    if not isinstance(section_1_last, DescriptionConverter):
        return True
    
    if section_1_last.line_count > 3:
        return True
    
    if not isinstance(section_2_first, DescriptionConverter):
        return True
    
    if section_1_last.indent+1 == section_2_first.indent:
        return False
    
    return True

class SectionConverter:
    """
    Converter class for converting a section.
    
    Attributes
    ----------
    parts : `list` of `Any`
        The parts of the section to render.
    """
    __slots__ = ('parts', )
    def __new__(cls, section, builder_context):
        """
        Creates a new section converter.
        
        Parameters
        ----------
        section : `tuple` ((`str` or `None`), `list` of `Any`)
            The section to represent.
        builder_context : ``BuilderContext``
            Context to define some building details.
        
        Returns
        -------
        self : ``SectionConverter``
        """
        section_name, section_parts = section
        parts = [
            *SectionTitleConverter(section_name),
            *sub_section_converter(section_parts, 0, builder_context.max_line_length, builder_context)
                ]
        
        self = object.__new__(cls)
        self.parts = parts
        return self
    
    @property
    def character_count(self):
        """
        Returns the section's characters' total length.
        
        Returns
        -------
        length : `int`
        """
        length = 0
        last_part = None
        for part in self.parts:
            if should_insert_linebreak(last_part, part):
                length +=1
            
            length += part.character_count
            last_part = part
        
        return length
    
    @property
    def line_count(self):
        """
        Returns the total line count of the section.
        
        Returns
        -------
        length : `int`
        """
        length = 0
        last_part = None
        for part in self.parts:
            if should_insert_linebreak(last_part, part):
                length +=1
            
            length += part.line_count
            last_part = part
        
        return length
    
    def _do_break(self, number_of_parts):
        """
        Breaks the section at the given part index.
        
        Parameters
        ----------
        number_of_parts : `int`
            The number of parts, what the first section should contain.
        
        Returns
        -------
        section_1 : ``SectionConverter``
        section_2 : ``SectionConverter``
        """
        parts = self.parts
        section_1 = object.__new__(type(self))
        section_1.parts = parts[:number_of_parts]
        section_2 = object.__new__(type(self))
        section_2.parts = parts[number_of_parts:]
        return section_1, section_2
    
    def _do_break_middle_chared(self, middle_part_index, middle_char_limit):
        """
        Breaks the section at the given part index, and tries to break the middle element based on the given middle
        char limit as well.
        
        Parameters
        ----------
        middle_part_index : `int`
            The number of parts, what the first section should contain besides the middle element.
        middle_char_limit : `int`
            The number of chars, where the middle part should be broken.
        
        Returns
        -------
        sections : `None` or `tuple` (``SectionConverter``, ``SectionConverter``)
            Returns `None`, if the middle part cannot be broken correctly.
        """
        parts = self.parts
        if len(parts) <= middle_part_index:
            return None
        
        middle_part = parts[middle_part_index]
        middle_broken = middle_part.do_break(middle_char_limit)
        if middle_broken is None:
            return None
        
        section_1 = object.__new__(type(self))
        section_1.parts = [*parts[:middle_part_index], middle_broken[0]]
        if not section_1._test_break():
            return None
        
        section_2 = object.__new__(type(self))
        section_2.parts = [middle_broken[1], *parts[middle_part_index+1:]]
        if not section_2._test_break():
            return None
        
        if not should_accept_section_break(section_1, section_2):
            return None
        
        return section_1, section_2
    
    def _test_break(self):
        """
        Returns whether this section shard should be rendered alone.
        
        Returns
        -------
        passed : `bool`
        """
        parts = self.parts
        if len(parts) == 0:
            return False
        
        if isinstance(parts[0], SectionTitleConverter):
            min_line_count = 7
        else:
            min_line_count = 5
        
        if self.line_count < min_line_count:
            return False
        
        return True
    
    def do_break(self, number_of_chars):
        """
        Breaks the section to parts an returns the most optimal case for the given max chars.
        
        Parameters
        ----------
        number_of_chars : `int`
            The maximal amount of chars, what the first section should contain.
        
        Returns
        -------
        best_fit : `None` or `tuple` (``SectionConverter``, ``SectionConverter``)
            The best fitting 2 section shard, if there is any optimal case. If there is non, returns `None` instead.
        """
        best_fit = None
        for break_point in range(1, len(self.parts)):
            sections = self._do_break(break_point)
            leftover_chars = number_of_chars - sections[0].character_count
            if leftover_chars < 0:
                break
            
            sections_middled = self._do_break_middle_chared(break_point, leftover_chars)
            if (sections_middled is not None):
                best_fit = sections_middled
                continue
            
            section_1, section_2 = sections
            if not section_1._test_break():
                continue
            
            if not section_2._test_break():
                continue
            
            if not should_accept_section_break(section_1, section_2):
                continue
            
            best_fit = sections
            continue
        
        if best_fit is None:
            best_fit = self._do_break_middle_chared(0, number_of_chars)
        
        return best_fit
    
    def render_to(self, to_extend):
        """
        Renders the section to the given `list`.
        
        Parameters
        ----------
        to_extend : `list` of `str`
            A list to what the section should yield it's lines.
        """
        last_part = None
        for part in self.parts:
            if should_insert_linebreak(last_part, part):
                to_extend.append('\n')
            
            part.render_to(to_extend)
            last_part = part
    
    def __repr__(self):
        """Returns the description's representation."""
        to_extend = [self.__class__.__name__, ':\n']
        self.render_to(to_extend)
        to_extend.append('<<END>>')
        return ''.join(to_extend)

def serialize_docs_embed_sized(docs):
    """
    Serializes the docs and returns a `list` of `str`, where each element has maximal length of the size of an embed
    description.
    
    Parameters
    ----------
    docs : ``DocString``
        The docstring to serialize.
    
    Returns
    -------
    result_chunks : `list` of `str`
    """
    result_chunks = []
    actual_chunk = []
    actual_chunk_size = 0
    
    for section in docs.sections:
        section = SectionConverter(section, EMBED_SIZED_BUILDER_CONTEXT)
        section_character_count = section.character_count
        
        limit = EMBED_DESCRIPTION_MAX_SIZE
        if actual_chunk:
            limit -= 1
        
        if actual_chunk_size+section_character_count <= limit:
            if actual_chunk:
                actual_chunk.append('\n')
                actual_chunk_size += 1
            
            section.render_to(actual_chunk)
            actual_chunk_size += section_character_count
            continue
        
        leftover = limit-actual_chunk_size-1
        sections = section.do_break(leftover)
        if sections is None:
            result_chunks.append(''.join(actual_chunk))
            actual_chunk.clear()
        else:
            section_1, section = sections
            if actual_chunk:
                actual_chunk.append('\n')
            
            section_1.render_to(actual_chunk)
            result_chunks.append(''.join(actual_chunk))
            actual_chunk.clear()
            section_character_count = section.character_count
        
        while section_character_count > EMBED_DESCRIPTION_MAX_SIZE:
            sections = section.do_break(EMBED_DESCRIPTION_MAX_SIZE)
            assert sections is not None, f'Cannot break down section: {section!r}'
            
            section_1, section = sections
            section_1.render_to(actual_chunk)
            result_chunks.append(''.join(actual_chunk))
            actual_chunk.clear()
            
            section_character_count = section.character_count
            continue
        
        section.render_to(actual_chunk)
        actual_chunk_size = section_character_count
        continue
    
    if actual_chunk:
        result_chunks.append(''.join(actual_chunk))
    
    return result_chunks

def serialize_docs(docs):
    """
    Serializes the given docs to one big string.
    
    Parameters
    ----------
    docs : ``DocString``
        The docstring to serialize.
    
    Returns
    -------
    result : `str`
    """
    result = []
    
    for section in docs.sections:
        SectionConverter(section, TEXT_BUILDER_CONTEXT).render_to(result)
    
    return ''.join(result)
    
def serialize_docs_source_text(docs):
    """
    Serializes the given docs to one big string with graves, like the source one.
    
    Parameters
    ----------
    docs : ``DocString``
        The docstring to serialize.
    
    Returns
    -------
    result : `str`
    """
    result = []
    
    for section in docs.sections:
        SectionConverter(section, TEXT_SOURCE_BUILDER_CONTEXT).render_to(result)
    
    return ''.join(result)

def generate_preview_for(docs, length=200, min_cut_off=20, end='...'):
    """
    Generates preview for the given docs.
    
    Parameters
    ----------
    docs : ``DocString``
        The docstring to serialize.
    length : `int`, Optional.
        The maximal length of the generated preview. Defaults to `200`.
    min_cut_off : `int`
        The minimal length after with new section part will not be added to the preview.
    end : `str`, Optional
        Ending string, what is added to the preview if a section would be cut in half. Defaults to `'...`'.
    
    Returns
    -------
    preview : `str`
    """
    min_cut_off = min_cut_off+len(end)+2
    
    name, sub_section = docs.sections[0]
    if name is not None:
        return None
    
    generated_parts = []
    for section_part in sub_section:
        # Skip Attribute description from previews.
        if type(section_part) is GravedAttributeDescription:
            continue
        
        if type(section_part) is not GravedDescription:
            break
        
        preview_part = preview_string(section_part.content)
        if preview_part is None:
            break
        
        if DO_NOT_PREVIEW_RP.fullmatch(preview_part) is not None:
            break
        
        length -= len(preview_part)
        if length < 0:
            preview_part = preview_part[:length-len(end)]
            generated_parts.append(preview_part)
            generated_parts.append(end)
            break
        
        generated_parts.append(preview_part)
        if length < min_cut_off:
            break
        
        length -= 2
        generated_parts.append('\n\n')
        continue
    
    if generated_parts:
        if generated_parts[-1] == '\n\n':
            del generated_parts[-1]
        preview = ''.join(generated_parts)
    else:
        preview = None
    
    return preview



# -*- coding: utf-8 -*-
__all__ = ('PermissionOverwrite', 'Role', 'cr_p_overwrite_object', 'cr_p_role_object', )

from ..backend.utils import DOCS_ENABLED

from ..env import API_VERSION

from .bases import DiscordEntity
from .client_core import ROLES
from .utils import random_id, DATETIME_FORMAT_CODE
from .color import Color
from .permission import Permission, PERMISSION_NONE
from .user import create_partial_user
from .preconverters import preconvert_snowflake, preconvert_str, preconvert_color, preconvert_int, preconvert_bool, \
    preconvert_flag
from .preinstanced import RoleManagerType

from . import rate_limit as module_rate_limit, user as module_user

create_partial_integration = NotImplemented

ROLE_MANAGER_TYPE_NONE        = RoleManagerType.none
ROLE_MANAGER_TYPE_UNSET       = RoleManagerType.unset
ROLE_MANAGER_TYPE_UNKNOWN     = RoleManagerType.unknown
ROLE_MANAGER_TYPE_BOT         = RoleManagerType.bot
ROLE_MANAGER_TYPE_BOOSTER     = RoleManagerType.booster
ROLE_MANAGER_TYPE_INTEGRATION = RoleManagerType.integration

if API_VERSION in (6, 7):
    PERMISSION_KEY = 'permissions_new'
    PERMISSION_ALLOW_KEY = 'allow_new'
    PERMISSION_DENY_KEY = 'deny_new'
    
    PERM_OW_TYPE_ROLE = 'role'
    PERM_OW_TYPE_USER = 'member'
    
    def get_perm_ow_key_value(data):
        """
        Returns the permission overwrite's type's value.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received permission overwrite data.
        
        Returns
        -------
        type_value : `str`
        """
        return data['type']
else:
    PERMISSION_KEY = 'permissions'
    PERMISSION_ALLOW_KEY = 'allow'
    PERMISSION_DENY_KEY = 'deny'
    
    PERM_OW_TYPE_ROLE = 0
    PERM_OW_TYPE_USER = 1
    
    def get_perm_ow_key_value(data):
        """
        Returns the permission overwrite's type's value.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received permission overwrite data.
        
        Returns
        -------
        type_value : `int`
        """
        return int(data['type'])


def create_partial_role(role_id):
    """
    Creates a partial role from the given `role_id`. If the role already exists returns that instead.
    
    Parameters
    ----------
    role_id : `int`
        The unique identifier number of the role.
    
    Returns
    -------
    role : ``Role``
    """
    try:
        return ROLES[role_id]
    except KeyError:
        pass
    
    role = object.__new__(Role)
    role.id = role_id
    
    role.color = Color()
    role.guild = None
    role.separated = False
    # id is set up
    role.manager_type = ROLE_MANAGER_TYPE_NONE
    role.manager_id = 0
    role.mentionable = False
    role.name = ''
    role.permissions = PERMISSION_NONE
    role.position = 1 # 0 is default role, so we go for 1
    
    ROLES[role_id] = role
    
    return role


class Role(DiscordEntity, immortal=True):
    """
    Represents a Discord guild's role.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the role.
    color : ``Color``
        The role's color. If the color equals to ``Color(0)``, then it is ignored meanwhile calculating towards a
        user's display color.
    guild : ``Guild`` or `None`
        The guild of the role. If the role is partial, including already deleted roles, then it's `.guild` is set to
        `None`.
    separated : `bool`
        Users show up in separated groups by their highest `separated` role.
    manager_id : `int`
        If the role is managed, then it's manager's id if applicable. Defaults to `0`.
    manager_type : `RoleManagerType`
        But what type of entity is the role managed.
    mentionable : `bool`
        Whether the role can be mentioned.
    name : `str`
        The name of the role. Can be empty string, if the role has no known name.
    permissions : ``Permission``
        The permissions of the users having the role.
    position : `int`
        The role's position.
    """
    __slots__ = ('color', 'guild', 'separated', 'manager_id', 'manager_type', 'mentionable', 'name', 'permissions',
        'position',)
    
    def __new__(cls, data, guild):
        """
        Creates a role from the given `data` at the given `guild`. If the role already exists and is not partial, then
        returns it. However it is partial, then updates it as well.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Role data received from Discord.
        guild : ``Guild``
            The owner guild of the role.
        
        Returns
        -------
        role : ``Role``
        """
        role_id = int(data['id'])
        try:
            role = ROLES[role_id]
            update = (role.guild is None)
        except KeyError:
            role = object.__new__(cls)
            role.id = role_id
            update = True
            ROLES[role_id] = role
        
        if update:
            
            guild.roles[role.id] = role
            role.guild = guild
            
            role.name = data['name']
            
            role.position = data.get('position', 0)
            
            role.color = Color(data.get('color', 0))
            
            role.permissions = Permission(data[PERMISSION_KEY])
            
            role.separated = data.get('hoist', False)
            
            if data.get('managed', False):
                role._set_managed(data)
            else:
                role.manager_id = 0
                role.manager_type = ROLE_MANAGER_TYPE_NONE
            
            role.mentionable = data.get('mentionable', False)
        
        return role
    
    @classmethod
    def precreate(cls, role_id, **kwargs):
        """
        Precreates a role by creating a partial one with the given parameters. When the role is loaded, then the
        precreated one will be picked up. However if the role is precreated when it already exists, then the existing
        one is picked up and is updated by the given parameters only if it is partial.
        
        Parameters
        ----------
        role_id : `int` or `str`
            The role's id.
        **kwargs : keyword arguments
            Additional predefined attributes for the role.
        
        Other parameters
        ----------------
        name : `str`
            The role's ``.name``.
        manager_id : `None`, `int` or `str`
            The role's manager's id.
        manager_type : `RoleManagerType`
            The role's ``.manager_type``.
        mentionable : `bool`
            The role's ``.mentionable``.
        separated : `bool`
            The role's ``.separated``.
        position : `int`
            The role's ``.position``.
        permissions : `int` or ``Permission``
            The role's ``.permissions``.
        color : `int` or ``Color``
            The role's ``.color``.
        
        Returns
        -------
        role : ``Role``
        
        Raises
        ------
        TypeError
            If any argument's type is bad or if unexpected argument is passed.
        ValueError
            If an argument's type is good, but it's value is unacceptable.
        """
        role_id = preconvert_snowflake(role_id, 'role_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 32)
                processable.append(('name',name))
            
            for key in ('mentionable', 'separated',):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    pass
                else:
                    value = preconvert_bool(value, key)
                    processable.append((key, value))
            
            try:
                position = kwargs.pop('position')
            except KeyError:
                pass
            else:
                position = preconvert_int(position, 'position', 0, 250)
                processable.append(('position', position))
            
            try:
                permissions = kwargs.pop('permissions')
            except KeyError:
                pass
            else:
                permissions = preconvert_flag(permissions, 'permissions', Permission)
                processable.append(('permissions', permissions))
            
            try:
                color = kwargs.pop('color')
            except KeyError:
                pass
            else:
                color = preconvert_color(color)
                processable.append(('color', color))
            
            try:
                manager_type = kwargs.pop('manager_type')
            except KeyError:
                try:
                    manager_id = kwargs.pop('manager_id')
                except KeyError:
                    pass
                else:
                    raise ValueError(f'`manager_type` was not given, meanwhile `manager_id` was. Received values: '
                        f'manager_id: {manager_id!r}')
            else:
                manager_type_type = manager_type.__class__
                if manager_type_type is not RoleManagerType:
                    raise TypeError(f'`manager_type` can be given as `{RoleManagerType.__class__}` instance, got '
                        f'{manager_type_type.__name__}')
                
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            role = ROLES[role_id]
        except KeyError:
            role = object.__new__(cls)
            role.id = role_id
            
            role.color = Color()
            role.guild = None
            role.separated = False
            role.manager_type = ROLE_MANAGER_TYPE_NONE
            role.mentionable = False
            role.name = ''
            role.permissions = PERMISSION_NONE
            role.position = 1
            ROLES[role_id] = role
        else:
            if (role.guild is not None):
                return role
        
        if (processable is not None):
            for item in processable:
                setattr(role, *item)
        
        return role
    
    def _set_managed(self, data):
        """
        Called when the role's manager_type is `ROLE_MANAGER_TYPE_UNSET` for trying to set the role's ``.manager_type``
        and ``.manager_id``.
        
        data : `dict` of (`str`, `Any`) items
            Role data received from Discord.
        """
        while True:
            try:
                tags = data['tags']
            except KeyError:
                manager_id = 0
                manager_type = ROLE_MANAGER_TYPE_UNSET
                break
            
            try:
                manager_id = tags['bot']
            except KeyError:
                pass
            else:
                manager_id = int(manager_id)
                manager_type = ROLE_MANAGER_TYPE_BOT
                break
            
            if 'premium_subscriber' in tags:
                manager_id = 0
                manager_type = ROLE_MANAGER_TYPE_BOOSTER
                break
            
            try:
                manager_id = data['integration']
            except KeyError:
                pass
            else:
                manager_id = int(manager_id)
                manager_type = ROLE_MANAGER_TYPE_INTEGRATION
                break
            
            manager_id = 0
            manager_type = ROLE_MANAGER_TYPE_UNKNOWN
            break
        
        self.manager_id = manager_id
        self.manager_type = manager_type
        
    def _update_no_return(self, data):
        """
        Updates the role with the given `data` with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Role data received from Discord.
        """
        guild = self.guild
        if guild is None:
            return
        
        clear_permission_cache = False
        
        position = data['position']
        if self.position != position:
            self.position = position
            clear_permission_cache = True
        
        self.name = data['name']
        
        permissions = Permission(data[PERMISSION_KEY])
        if self.permissions != permissions:
            self.permissions = permissions
            clear_permission_cache = True
        
        self.color = Color(data['color'])
        self.separated = data['hoist']
        
        if self.manager_type is ROLE_MANAGER_TYPE_UNSET:
            self._set_managed(data)
        
        self.mentionable = data['mentionable']
        
        if clear_permission_cache:
            guild._invalidate_perm_cache()
        
    def __str__(self):
        """Returns the role"s name or `'Partial'` if it has non."""
        name = self.name
        if not name:
            name = 'Partial'
        return name
    
    def __repr__(self):
        """Returns the role's representation."""
        return f'<{self.__class__.__name__} name={self.name!r}, id={self.id}>'
    
    def _update(self, data):
        """
        Updates the role with the given `data` and returns it's overwritten attributes as a `dict` with a
        `attribute-name` - `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Role data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +---------------+-------------------+
        | Keys          | Values            |
        +===============+===================+
        | color         | ``Color``         |
        +---------------+-------------------+
        | managed       | `bool`            |
        +---------------+-------------------+
        | mentionable   | `bool`            |
        +---------------+-------------------+
        | name          | `str`             |
        +---------------+-------------------+
        | permissions   | ``Permission``    |
        +---------------+-------------------+
        | position      | `int`             |
        +---------------+-------------------+
        | separated     | `bool`            |
        +---------------+-------------------+
        """
        old_attributes = {}
        guild = self.guild
        if guild is None:
            return old_attributes
        
        clear_permission_cache = False
        
        position = data['position']
        if self.position != position:
            old_attributes['position'] = self.position
            self.position = position
            clear_permission_cache = True
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        permissions = Permission(data[PERMISSION_KEY])
        if self.permissions != permissions:
            old_attributes['permissions'] = self.permissions
            self.permissions = permissions
            clear_permission_cache = True
        
        color = data['color']
        if self.color != color:
            old_attributes['color'] = self.color
            self.color = Color(color)
        
        separated = data['hoist']
        if self.separated != separated:
            old_attributes['separated'] = self.separated
            self.separated = separated
        
        if self.manager_type is ROLE_MANAGER_TYPE_UNSET:
            self._set_managed(data)
        
        managed = data['managed']
        if self.managed != managed:
            old_attributes['managed'] = self.managed
            self.managed = managed
        
        mentionable = data['mentionable']
        if self.mentionable != mentionable:
            old_attributes['mentionable'] = self.mentionable
            self.mentionable = mentionable
        
        if clear_permission_cache:
            guild._invalidate_perm_cache()
        
        return old_attributes
    
    def _delete(self):
        """
        Removes the role's references.
        
        Used when the role is deleted.
        """
        guild = self.guild
        if guild is None:
            return # already deleted
        
        self.guild = None
        
        try:
            del guild.roles[self.id]
        except KeyError:
            pass
        
        guild._cache_perm = None
        for channel in guild.channels.values():
            channel._cache_perm = None
        
        for user in guild.users.values():
            try:
                profile = user.guild_profiles[guild]
            except KeyError:
                # the user has no ``GuildProfile``, it supposed to be impossible
                continue
            
            roles = profile.roles
            if roles is None:
                continue
            
            try:
                roles.remove(self)
            except ValueError:
                pass
    
    @property
    def is_default(self):
        """
        Returns whether the role is the default role at it's respective guild.
        
        Returns
        -------
        is_default : `bool`
        """
        return (self.position == 0)
    
    @property
    def mention(self):
        """
        The role's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<@&{self.id}>'
    
    def __format__(self, code):
        """
        Formats the role in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        role : `str`
        
        Examples
        --------
        ```py
        >>>> from hata import Role, now_as_id
        >>>> role = Role.precreate(now_as_id(), name='admiral-general')
        >>>> role
        <Role name='admiral-general', id=725333995067277312>
        >>>> # no code stands for str(role).
        >>>> f'{role}'
        'admiral-general'
        >>>> # 'm' stands for mention.
        >>>> f'{role:m}'
        '<@&725333995067277312>'
        >>>> # 'c' stands for created at.
        >>>> f'{role:c}'
        '2020.06.24-12:58:20'
        ```
        """
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        if code == 'c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    @property
    def users(self):
        """
        Returns the users who have the role.
        
        Returns
        -------
        users : `list` of (``User`` or ``Client``) instances
        """
        guild = self.guild
        if guild is None:
            users = []
        elif self.position == 0:
            users = list(guild.users.values())
        else:
            users = []
            for user in guild.users.values():
                try:
                    profile = user.guild_profiles[guild]
                except KeyError:
                    # should not happen
                    continue
                
                roles = profile.roles
                if roles is None:
                    continue
                
                if self not in user.roles:
                    continue
                
                users.append(user)
        
        return users
    
    @property
    def managed(self):
        """
        Returns whether the role is managed.
        
        Returns
        -------
        managed: `bool`
        """
        return (self.manager_type is not ROLE_MANAGER_TYPE_NONE)
    
    @property
    def manager(self):
        """
        Returns the manager entity of the role if applicable.
        
        Returns
        -------
        manager : `None`, ``User``, ``Client`` or ``Integration``
        """
        manager_type = self.manager_type
        if manager_type is ROLE_MANAGER_TYPE_NONE:
            manager = None
        
        elif manager_type is ROLE_MANAGER_TYPE_BOT:
            manager = create_partial_user(self.manager_id)
            
            # `create_partial_user` sets newly created users' `.is_bot` attribute as `False`.
            if not manager.is_bot :
                manager.is_bot = True
        
        elif manager_type is ROLE_MANAGER_TYPE_INTEGRATION:
            manager = create_partial_integration(self.manager_id, role=self)
        
        else:
            manager = None
        
        return manager
    
    @property
    def partial(self):
        """
        Returns whether the role is partial.
        
        Partial roles have `.guild` set as ``None``.
        
        Returns
        -------
        partial : `bool`
        """
        return (self.guild is None)
    
    def _get_colour(self):
        return self.color
    
    def _set_colour(self, color):
        self.color = color
    
    colour = property(_get_colour, _set_colour)
    if DOCS_ENABLED:
        colour.__doc__ = """Alias of ``.color``."""
    
    def __gt__(self, other):
        """Returns whether this role's position is higher than the other's."""
        if type(self) is type(other):
            if self.position > other.position:
                return True
            
            if self.position == other.position:
                if self.id > other.id:
                    return True
        
            return False
        
        return NotImplemented
    
    def __ge__(self, other):
        """Returns whether this role's position is higher or equal to the other's."""
        if type(self) is type(other):
            if self.position > other.position:
                return True
            
            if self.position == other.position:
                if self.id >= other.id:
                    return True
            
            return False
        
        return NotImplemented
    
    def __le__(self, other):
        """Returns whether this role's position is less or equal to the other's."""
        if type(self) is type(other):
            if self.position < other.position:
                return True
            
            if self.position == other.position:
                if self.id <= other.id:
                    return True
            
            return False
        
        return NotImplemented
    
    def __lt__(self, other):
        """Returns whether this role's position is less than the other's."""
        if type(self) is type(other):
            if self.position < other.position:
                return True
            
            if self.position == other.position:
                if self.id < other.id:
                    return True
            
            return False
        
        return NotImplemented


class PermissionOverwrite:
    """
    Represents a permission overwrite of a guild channel.
    
    Attributes
    ----------
    allow : ``Permission``
        The allowed permissions by the overwrite.
    deny : ``Permission``
        The denied permission by the overwrite.
    target_role : `None`, ``Role`` or ``Unknown``
        The target role entity of the overwrite if applicable. Defaults to `None`.
    target_user_id : `int`
        The target user id of the overwrite if applicable. Defaults to `0`.
    """
    __slots__ = ('allow', 'deny', 'target_role', 'target_user_id')
    
    def __init__(self, data):
        """
        Creates a permission overwrite from the given data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received permission overwrite data.
        """
        id_ = int(data['id'])
        if get_perm_ow_key_value(data) == PERM_OW_TYPE_ROLE:
            target_role = create_partial_role(id_)
            target_user_id = 0
        else:
            target_role = None
            target_user_id = id_
        
        self.target_role = target_role
        self.target_user_id = target_user_id
        self.allow = Permission(data[PERMISSION_ALLOW_KEY])
        self.deny = Permission(data[PERMISSION_DENY_KEY])
    
    @property
    def target(self):
        """
        Returns the target entity of the overwrite.
        
        Returns
        -------
        target : ``Client``, ``User`` or ``Role``
        """
        target = self.target_role
        if target is None:
            target = create_partial_user(self.target_user_id)
        
        return target
    
    @classmethod
    def custom(cls, target, allow, deny):
        """
        Creates an overwrite object with the given parameters.
        
        Parameters
        ----------
        target : ``Role`` or ``UserBase`` instance
            The target entity of the overwrite.
        allow : `int`
            The allowed permissions by the overwrite.
        deny : `int`
            The denied permission by the overwrite.
        
        Returns
        -------
        self : ``PermissionOverwrite``
        """
        if type(target) is Role:
            target_role = target
            target_user_id = 0
        else:
            target_role = None
            target_user_id = target.id
        
        self = object.__new__(cls)
        self.target_role = target_role
        self.target_user_id = target_user_id
        self.allow = Permission(allow)
        self.deny = Permission(deny)
        return self
    
    def __hash__(self):
        """Returns the permission overwrite's hash."""
        return self.target.id^self.allow^self.deny
    
    def __repr__(self):
        """Returns the permission overwrite's representation."""
        return f'<{self.__class__.__name__} target={self.target!r}>'
    
    def keys(self):
        """
        Yields the permissions' names.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
            Permissions' respective name.
        """
        yield from Permission.__keys__.keys()
    
    __iter__ = keys
    
    def values(self):
        """
        Yields position by position each permission's state. `+1` is yielded if the permission is enabled, `-1` if
        disabled and `0` if neither.
        
        This method is a generator.
        
        Yields
        ------
        state : `int`
            The permission's state.
            
            Can be one of the following:
            
            +-----------+-------+
            | Name      | Value |
            +===========+=======+
            | Enabled   | +1    |
            +-----------+-------+
            | None      | 0     |
            +-----------+-------+
            | Disabled  | -1    |
            +-----------+-------+
        """
        allow = self.allow
        deny = self.deny
        for index in Permission.__keys__.values():
            if (allow>>index)&1:
                state = +1
            elif (deny>>index)&1:
                state = -1
            else:
                state = 0
            
            yield state
    
    def items(self):
        """
        Yields the permission overwrite's items. What includes their name and their state. As state `+1` is yielded
        if the permission is enabled, `-1` if disabled and `0` if neither.
        
        This method is a generator.
        
        Yields
        -------
        name : str`
            Permissions' respective name.
        state : `int`
            The permission's state.
            
            Can be one of the following:
            
            +-----------+-------+
            | Name      | Value |
            +===========+=======+
            | Enabled   | +1    |
            +-----------+-------+
            | None      | 0     |
            +-----------+-------+
            | Disabled  | -1    |
            +-----------+-------+
        """
        allow = self.allow
        deny = self.deny
        for key, index in Permission.__keys__.items():
            if (allow>>index)&1:
                state = +1
            elif (deny>>index)&1:
                state = -1
            else:
                state = 0
            
            yield key, state
    
    def __getitem__(self, key):
        """Returns the permission's state for the given permission name."""
        index = Permission.__keys__[key]
        if (self.allow>>index)&1:
            state = +1
        elif (self.deny>>index)&1:
            state = -1
        else:
            state = 0
        
        return state
    
    @property
    def type(self):
        if type(self.target) is Role:
            type_ = PERM_OW_TYPE_ROLE
        else:
            type_ = PERM_OW_TYPE_USER
        return type_
    
    if DOCS_ENABLED:
        if API_VERSION in (6, 7):
            type.__doc__ = (
        """
        Returns the Discord side identifier value permission overwrite.
        
        Returns
        -------
        type_ : `str`
            Can be either `'role'` or `'member'`
        """)
        else:
            type.__doc__ = (
        """
        Returns the Discord side identifier value permission overwrite.
        
        Returns
        -------
        type_ : `int`
            Can be either 0` or `1`
        """)
    
    @property
    def id(self):
        """
        Returns the permission overwrite's target's id.
        
        Returns
        -------
        id_ : `int`
        """
        target_role = self.target_role
        if target_role is None:
            id_ = self.target_user_id
        else:
            id_ = target_role.id
        
        return id_
    
    def __lt__(self, other):
        """Returns whether is this permission overwrite is at lower position in ordering than the order."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_target_role = self.target_role
        if self_target_role is None:
            self_target_type = 0
            self_target_id = self.target_user_id
        else:
            self_target_type = 1
            self_target_id = self_target_role.id
        
        other_target_role = other.target_role
        if other_target_role is None:
            other_target_type = 0
            other_target_id = other.target_user_id
        else:
            other_target_type = 1
            other_target_id = other_target_role.id
        
        if self_target_type < other_target_type:
            return True
        
        if self_target_type == other_target_type:
            if self_target_id < other_target_id:
                return True
            
            return False
        
        return False
    
    def __eq__(self, other):
        """Returns whether is this permission overwrite is same as the other."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.allow != other.allow:
            return False
        
        if self.deny != other.deny:
            return False
        
        self_target_role = self.target_role
        if self_target_role is None:
            self_target_id = self.target_user_id
        else:
            self_target_id = self_target_role.id
        
        other_target_role = other.target_role
        if other_target_role is None:
            other_target_id = other.target_user_id
        else:
            other_target_id = other_target_role.id
        
        if self_target_id != other_target_id:
            return False
        
        return True
    
    def __gt__(self, other):
        """Returns whether is this permission overwrite is at greater position in ordering than the order."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_target_role = self.target_role
        if self_target_role is None:
            self_target_type = 0
            self_target_id = self.target_user_id
        else:
            self_target_type = 1
            self_target_id = self_target_role.id
            
        other_target_role = other.target_role
        if other_target_role is None:
            other_target_type = 0
            other_target_id = other.target_user_id
        else:
            other_target_type = 1
            other_target_id = other_target_role.id
        
        if self_target_type > other_target_type:
            return True
        
        if self_target_type == other_target_type:
            if self_target_id > other_target_id:
                return True
            
            return False
        
        return False

def cr_p_role_object(name, id_=None, color=Color(), separated=False, position=0, permissions=Permission(),
        managed=False, mentionable=False):
    """
    Creates a json serializable object representing a ``Role``.
    
    Parameters
    ----------
    name : `str`
        The name of the role.
    id_ : `None` or `int`,`optional
        The role's unique identifier number. If given as `None`, then a random `id` will be generated.
    color : ``Color``, Optional
        The role's color. Defaults to `Color(0)`
    separated : `bool`
        Users show up in separated groups by their highest `separated` role. Defaults to `False`.
    position : `int`
        The role's position at the guild. Defaults to `0`.
    permissions : ``Permission``
        The permissions of the users having the role.
    managed : `bool`
        Whether the role is managed by an integration.
    mentionable : `bool`
        Whether the role can be mentioned.
    
    Returns
    -------
    role_data : `dict` of (`str`, `Any`) items
    """
    if id_ is None:
        id_ = random_id()
    
    return {
        'id'          : id_,
        'name'        : name,
        'color'       : color,
        'hoist'       : separated,
        'position'    : position,
        'permissions' : permissions,
        'managed'     : managed,
        'mentionable' : mentionable,
            }

def cr_p_overwrite_object(target, allow, deny):
    """
    Creates a json serializable object representing a ``PermissionOverwrite``(permission overwrite).
    
    Parameters
    ----------
    target : ``Client``, ``User`` or ``Role``
        The target entity of the overwrite.
        The allowed permissions by the overwrite.
    deny : `int`
        The denied permission by the overwrite.
    
    Returns
    -------
    permission_overwrite_data : `dict` of (`str`, `Any) items
    """
    return {
        'allow' : allow,
        'deny'  : deny,
        'id'    : target.id,
        'type'  : PERM_OW_TYPE_ROLE if type(target) is Role else PERM_OW_TYPE_USER,
            }

module_rate_limit.Role = Role
module_user.create_partial_role = create_partial_role

del module_rate_limit
del module_user



# -*- coding: utf-8 -*-
__all__ = ('Application', 'ApplicationExecutable', 'ApplicationSubEntity', 'EULA', 'Team', 'TeamMember',
    'ThirdPartySKU', )

from .bases import DiscordEntity, IconSlot, ICON_TYPE_NONE, FlagBase
from .http import URLS
from .user import ZEROUSER, User
from .client_core import TEAMS, EULAS, APPLICATIONS, USERS
from .preinstanced import TeamMembershipState
from .preconverters import preconvert_snowflake, preconvert_bool, preconvert_str, preconvert_flag

Client = NotImplemented

class ApplicationFlag(FlagBase):
    """
    Represents an application's flags.
    
    The implemented user flags are the following:
    
    +-----------------------------------+-------------------+
    | Respective name                   | Bitwise position  |
    +===================================+===================+
    | managed_emoji                     | 2                 |
    +-----------------------------------+-------------------+
    | group_dm_create                   | 4                 |
    +-----------------------------------+-------------------+
    | rpc_hash_connected                | 11                |
    +-----------------------------------+-------------------+
    | gateway_presence                  | 12                |
    +-----------------------------------+-------------------+
    | gateway_presence_limited          | 13                |
    +-----------------------------------+-------------------+
    | gateway_guild_members             | 14                |
    +-----------------------------------+-------------------+
    | gateway_guild_members_limited     | 15                |
    +-----------------------------------+-------------------+
    | verification_pending_guild_limit  | 16                |
    +-----------------------------------+-------------------+
    | embedded                          | 17                |
    +-----------------------------------+-------------------+
    """
    __keys__ = {
        'managed_emoji'                    :  2,
        'group_dm_create'                  :  4,
        'rpc_hash_connected'               : 11,
        'gateway_presence'                 : 12,
        'gateway_presence_limited'         : 13,
        'gateway_guild_members'            : 14,
        'gateway_guild_members_limited'    : 15,
        'verification_pending_guild_limit' : 16,
        'embedded'                         : 17,
            }

class Application(DiscordEntity, immortal=True):
    """
    Represents a Discord application with all of it's spice.
    
    When a ``Client`` is created, it starts it's life with an empty application by default. However when the client
    logs in, it's application is requested, but it can be updated by ``Client.update_application_info`` anytime.
    
    Attributes
    ----------
    aliases : `None` or `list` of `str`
        Aliases of the application's name. Defaults to `None`.
    bot_public : `bool`.
        Whether not only the application's owner can join the application's bot to guilds. Defaults to `False`
    bot_require_code_grant : `bool`
        Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        Defaults to `False`.
    cover_hash : `int`
        The application's store cover image's hash in `uint128`. If the application is sold at Discord, this image
        will be used at the store.
    cover_type : ``IconType``
        The application's store cover image's type.
    description : `str`
        The description of the application. Defaults to empty string.
    developers : `None` or `list` of ``ApplicationSubEntity``
        A list of the application's games' developers. Defaults to `None`.
    eula_id : `int`
        The end-user license agreement's id of the application. Defaults to `0` if not applicable.
    executables : `None` or `list` of ``ApplicationExecutable``
        A list of the application's executables. Defaults to `None`.
    flags : ``ApplicationFlag``
        The application's public flags.
    guild_id : `int`
        If the application is a game sold on Discord, this field tells in which guild it is.
        Defaults to `0` if not applicable.
    hook : `bool`
        Defaults to `False`.
    icon_hash : `int`
        The application's icon's hash as `uint128`.
    icon_type : ``IconType``
        The application's icon's type.
    id : `int`
        The application's id. Defaults to `0`. Meanwhile set as `0`, hashing the application will raise `RuntimeError`.
    name : `str`
        The name of the application. Defaults to empty string.
    overlay : `bool`
        Defaults to `False`.
    overlay_compatibility_hook : `bool`
        Defaults to `False`.
    owner : ``User``, ``Client`` or ``Team``
        The application's owner. Defaults to `ZEROUSER`.
    primary_sku_id : `int`
        If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
        Defaults to `0`.
    privacy_policy_url : `str` or 7None`
        The url of the application's privacy policy. Defaults to `None`.
    publishers : `None` or `list` of ``ApplicationSubEntity``
        A list of the application's games' publishers. Defaults to `None`.
    rpc_origins : `None` or `list` of `str`
        A list of `rpc` origin urls, if `rpc` is enabled. Set as `None` if would be an empty list.
    slug : `str` or `None`
        If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        Defaults to `None`.
    splash_hash : `int`
        The application's splash image's hash as `uint128`.
    splash_type : ``IconType``
        The application's splash image's type.
    summary : `str`
        If this application is a game sold on Discord, this field will be the summary field for the store page of its
        primary sku. Defaults to empty string.
    terms_of_service_url : `str` or `None`
        The url of the application's terms of service. Defaults to `None`.
    third_party_skus : `None` or `list` of ``ThirdPartySKU``
         A list of the application's third party stock keeping units. Defaults to `None`.
    verify_key : `str`
        A base64 encoded key for the GameSDK's `GetTicket`. Defaults to empty string.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = ('aliases', 'bot_public', 'bot_require_code_grant', 'description', 'developers', 'eula_id',
        'executables', 'flags', 'guild_id', 'hook', 'name', 'overlay', 'overlay_compatibility_hook', 'owner',
        'primary_sku_id', 'privacy_policy_url', 'publishers', 'rpc_origins', 'slug', 'summary', 'terms_of_service_url',
        'third_party_skus', 'verify_key')
    
    cover = IconSlot('cover', 'cover_image', URLS.application_cover_url, URLS.application_cover_url_as, add_updater=False)
    icon = IconSlot('icon', 'icon', URLS.application_icon_url, URLS.application_icon_url_as, add_updater=False)
    splash = IconSlot('splash', 'splash', None, None, add_updater=False)
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty application, with it's default attributes set.
        
        Returns
        -------
        self : ``Application``
            The created application.
        """
        self = object.__new__(cls)
        
        self.id = 0
        self.name = ''
        self.description = ''
        self.rpc_origins = None
        self.bot_public = False
        self.bot_require_code_grant = False
        self.owner = ZEROUSER
        self.summary = ''
        self.verify_key = ''
        self.guild_id = None
        self.primary_sku_id = 0
        self.slug = None
        self.developers = None
        self.hook = None
        self.publishers = None
        self.executables = None
        self.third_party_skus = None
        self.overlay = False
        self.overlay_compatibility_hook = False
        self.aliases = None
        self.eula_id = 0
        self.flags = ApplicationFlag()
        self.privacy_policy_url = None
        self.terms_of_service_url = None
        
        self.cover_hash = 0
        self.cover_type = ICON_TYPE_NONE
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        self.splash_hash = 0
        self.splash_type = ICON_TYPE_NONE
        
        return self
    
    def _create_update(self, data, ready_data):
        """
        Creates a new application if the given data refers to an other one. Updates the application and returns it.
        
        Parameters
        ----------
        data : or `dict` of (`str`, `Any`) items, Optional
            Application data received from Discord.
        ready_data : `bool`
            Whether the application data was received from a ready event.
        
        Returns
        -------
        self : ``Application``
            The created or updated application.
        """
        application_id = int(data['id'])
        
        if self.id == 0:
            try:
                self = APPLICATIONS[application_id]
            except KeyError:
                self.id = application_id
                APPLICATIONS[application_id] = self
        
        if ready_data:
            try:
                flags = data['flags']
            except KeyError:
                flags = ApplicationFlag()
            else:
                flags = ApplicationFlag(flags)
            self.flags = flags
        else:
            self._update_no_return(data, set_owner=True)
        
        return self
    
    def __new__(cls, data):
        """
        Creates a new application with the given data. If the application already exists, updates it and returns that
        instead.
        
        Parameters
        ----------
        data : or `dict` of (`str`, `Any`) items, Optional
            Application data received from Discord.
        
        Returns
        -------
        self : `˙Application``
        """
        application_id = int(data['id'])
        try:
            self = APPLICATIONS[application_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = application_id
            self.flags = ApplicationFlag()
            self._update_no_return(data, set_owner=True)
        else:
            self._update_no_return(data)
        
        return self
    
    def __hash__(self):
        """Returns the application's hash value."""
        id_ = self.id
        if id_:
            return id_
        
        raise TypeError(f'Cannot hash partial {self.__class__.__name__} object.')
    
    @property
    def partial(self):
        """
        Returns whether the application is partial.
        
        An application if partial, if it's id is set as `0`.
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    def __repr__(self):
        """Returns the application's representation"""
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        id_ = self.id
        if id_:
            result.append(' id=')
            result.append(repr(id_))
            result.append(', name=')
            result.append(repr(self.name))
        else:
            result.append(' partial')
        
        result.append('>')
        
        return ''.join(result)
    
    def __str__(self):
        """Returns the application's name."""
        return self.name
    
    def _update_no_return(self, data, set_owner=False):
        """
        Updates the application with the data received from Discord.
        
        Parameters
        ----------
        data : `None` or `dict` of (`str`, `Any`) items
            Application data received from Discord.
        set_owner : `bool`, Optional
            Whether the application's owner should be set from the given data. Defaults to `False`.
            Should be given as `True`, if the application is created or if it contains owner data.
        """
        self.name = data['name']
        
        self.description = data['description']
        
        rpc_origins = data.get('rpc_origins')
        if (rpc_origins is not None) and (not rpc_origins):
            rpc_origins = None
        
        self.rpc_origins = rpc_origins
        
        self.bot_public = data.get('bot_public', False)
        self.bot_require_code_grant = data.get('bot_require_code_grant', False)
        self.summary = data['summary']
        self.verify_key = data['verify_key']
        
        if set_owner:
            team_data = data.get('team')
            if team_data is None:
                owner_data = data.get('owner')
                if owner_data is None:
                    owner = ZEROUSER
                else:
                    owner = User(owner_data)
            else:
                owner = Team(team_data)
            
            self.owner = owner
        
        guild_id = data.get('guild_id')
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self.guild_id = guild_id
        
        primary_sku_id = data.get('primary_sku_id')
        if primary_sku_id is None:
            primary_sku_id = None
        else:
            primary_sku_id = int(primary_sku_id)
        self.primary_sku_id = primary_sku_id
        
        self.slug = data.get('slug', None)
        
        self._set_cover(data)
        self._set_icon(data)
        
        developers_data = data.get('developers')
        if (developers_data is None) or (not developers_data):
            developers = None
        else:
            developers = [ApplicationSubEntity(developer_data) for developer_data in developers_data]
        
        self.developers = developers
        
        self.hook = data.get('hook', False)
        
        publishers_data = data.get('publishers')
        if (publishers_data is None) or (not publishers_data):
            publishers = None
        else:
            publishers = [ApplicationSubEntity(publisher_data) for publisher_data in publishers_data]
        
        self.publishers = publishers
        
        executables_data = data.get('executables')
        if (executables_data is None) or (not executables_data):
            executables = None
        else:
            executables = [ApplicationExecutable(executable_data) for executable_data in executables_data]
        
        self.executables = executables
        
        self._set_splash(data)
        
        third_party_skus_data = data.get('third_party_skus')
        if (third_party_skus_data is None) or (not third_party_skus_data):
            third_party_skus = None
        else:
            third_party_skus = [ThirdPartySKU(third_party_sku_data) for third_party_sku_data in third_party_skus_data]
        
        self.third_party_skus = third_party_skus
        
        self.overlay = data.get('overlay', False)
        self.overlay_compatibility_hook = data.get('overlay_compatibility_hook', False)
        
        aliases = data.get('aliases')
        if (aliases is not None) and (not aliases):
            aliases = None
        
        self.aliases = aliases
        
        eula_id = data.get('eula_id')
        if eula_id is None:
            eula_id = 0
        else:
            eula_id = int(eula_id)
        
        self.eula_id = eula_id
        
        # Update data may not contain flags, so do not set if missing.
        try:
            flags = data['flags']
        except KeyError:
            pass
        else:
            self.flags = ApplicationFlag(flags)
        
        privacy_policy_url = data.get('privacy_policy_url')
        if (privacy_policy_url is not None) and (not privacy_policy_url):
            privacy_policy_url = None
        self.privacy_policy_url = privacy_policy_url
        
        terms_of_service_url = data.get('terms_of_service_url')
        if (terms_of_service_url is not None) and (not terms_of_service_url):
            terms_of_service_url = None
        self.terms_of_service_url = terms_of_service_url
        
    @classmethod
    def precreate(cls, application_id, **kwargs):
        """
        Precreates an application with the given parameters.
        
        Parameters
        ----------
        application_id : `int` or `str`
            The application's id.
        **kwargs : keyword arguments
            Additional predefined attributes for the application.
        
        Other parameters
        ----------------
        bot_public : `bool`, Optional
            Whether not only the application's owner can join the application's bot to guilds.
        bot_require_code_grant : `bool`, Optional
            Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        description : `str`, Optional
            The description of the application.
        flags : `int` or ``ApplicationFlag``, Optional
            The application's public flags. If not given as ``ApplicationFlag`` instance, then is converted to it.
        icon : `None`, ``Icon`` or `str`, Optional
            The application's icon.
            
            > Mutually exclusive with `icon_type` and `icon_hash`.
        icon_type : ``IconType``, Optional
            The application's icon's type.
            
            > Mutually exclusive with `icon`.
        icon_hash : `int`, Optional
            The application's icon's hash.
            
            > Mutually exclusive with `icon`.
        owner : ``User``, ``Client`` or ``Team``
            The application's owner. Defaults to `ZEROUSER`.
            
            This field cannot be given as `snowflake`, because it might represent both ``UserBase`` instances  and
            ``Team``-s as well.
        privacy_policy_url : `None` or `str`, Optional
            The url of the application's privacy policy.
        slug : `str` or `None`
            If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        splash : `None`, ``Icon`` or `str`, Optional
            The application's splash.
            
            > Mutually exclusive with `icon_type` and `icon_hash`.
        splash_type : ``IconType``, Optional
            The application's icon's type.
            
            > Mutually exclusive with `icon`.
        splash_hash : `int`, Optional
            The application's icon's hash.
            
            > Mutually exclusive with `icon`.
        
        summary : `str`, Optional
            If this application is a game sold on Discord, this field will be the summary field for the store page of
            its primary sku.
        terms_of_service_url : `str` or `None`, Optional
            The url of the application's terms of service.
        
        Returns
        -------
        application : `Application``
        
        Raises
        ------
        TypeError
            If any argument's type is bad or if unexpected argument is passed.
        ValueError
            If an argument's type is good, but it's value is unacceptable.
        """
        application_id = preconvert_snowflake(application_id, 'application_id')
        
        if kwargs:
            processable = []
            
            for key in ('bot_public', 'bot_require_code_grant'):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                value = preconvert_bool(value, key)
                processable.append((key, value))
            
            for key in ('description', 'summary'):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                value = preconvert_str(value, key, 0, 1024)
                processable.append((key, value))
            
            try:
                flags = kwargs.pop('flags')
            except KeyError:
                pass
            else:
                flags = preconvert_flag(flags, 'flags', ApplicationFlag)
                processable.append(('flags', flags))
            
            cls.icon.preconvert(kwargs, processable)
            
            try:
                owner = kwargs.pop('owner')
            except KeyError:
                pass
            else:
                if not isinstance(owner, (User, Client, Team)):
                    raise TypeError(f'`owner` can be given as {User.__name__}, {Client.__name__}  or as '
                        f'{Team.__name__} instance, got {owner.__class__.__name__}.')
                
                processable.append(('owner', owner))
            
            for key in ('slug', 'privacy_policy_url', 'terms_of_service_url'):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                if (value is None):
                    continue
                
                value = preconvert_str(value, key, 0, 1024)
                if not value:
                    continue
                
                processable.append((key, value))
            
            cls.splash.preconvert(kwargs, processable)
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}.')
            
        else:
            processable = None
        
        try:
            application = APPLICATIONS[application_id]
        except KeyError:
            application = cls._create_empty()
            application.id = application_id
            is_partial = True
        else:
            is_partial = application.partial
        
        if is_partial and (processable is not None):
            for item in processable:
                setattr(application, *item)
        
        return application


class Team(DiscordEntity, immortal=True):
    """
    Represents a Team on the Discord developer portal.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the team.
    icon_hash : `int`
        The team's icon's hash as `uint128`. Defaults to `0`.
    icon_type : `IconType`
        The team's icon's type.
    members : `list` of `TeamMember`
        The members of the team. Includes invited members as well.
    name : `str`
        The teams name.
    owner_id : `int`
        The team's owner's id.
    
    Notes
    -----
    Team objects support weakreferencing.
    """
    __slots__ = ('members', 'name', 'owner_id',)
    
    icon = IconSlot('icon', 'icon', URLS.team_icon_url, URLS.team_icon_url_as, add_updater = False)
    
    def __new__(cls, data):
        """
        Creates a new ``Team`` instance from the data received from Discord.
        
        Parameters
        ----------
        data : `None` or `dict` of (`str`, `Any`) items
            Team data received from Discord.
        
        Returns
        -------
        team : ``Team``
        """
        team_id = int(data['id'])
        try:
            team = TEAMS[team_id]
        except KeyError:
            team = object.__new__(cls)
            team.id = team_id
        
        #update every attribute
        team.name = data['name']
        
        team._set_icon(data)
        
        team.members = [TeamMember(team_member_data) for team_member_data in data['members']]
        team.owner_id = int(data['owner_user_id'])
        return team
    
    @property
    def owner(self):
        """
        Returns the team's owner.
        
        Returns
        -------
        owner : ``Client`` or ``User``
            Defaults to `ZEROUSER`.
        """
        owner_id = self.owner_id
        try:
            owner = USERS[owner_id]
        except KeyError:
            owner = ZEROUSER
        
        return owner
    
    @property
    def invited(self):
        """
        A list of the invited users to the team.
        
        Returns
        -------
        users : `list` of (``User`` or ``Client``) objects
        """
        target_state = TeamMembershipState.invited
        return [team_member.user for team_member in self.members if team_member.state is target_state]
    
    @property
    def accepted(self):
        """
        A list of the users, who accepted their invite to the team.
        
        Returns
        -------
        users : `list` of (``User`` or ``Client``) objects
        """
        target_state = TeamMembershipState.accepted
        return [team_member.user for team_member in self.members if team_member.state is target_state]
    
    def __str__(self):
        """Returns the team's name."""
        return self.name
    
    def __repr__(self):
        """Returns the team's representation."""
        return f'<{self.__class__.__name__} owner={self.owner.full_name}, total members={len(self.members)}>'

class TeamMember:
    """
    Represents a team member of a ``Team``.
    
    Attributes
    ----------
    permissions : `list` of `str`
        The permissions of the team member. Right now specific permissions are not supported, so the list has only
        one element : `'*'`, what represents all the permissions.
    state : ``TeamMembershipState``
        The state of the team member. A member can be invited or can have the invite already accepted.
    user : ``User`` or ``Client``
        The corresponding user account of the team member object.
    """
    __slots__ = ('permissions', 'state', 'user',)
    
    def __init__(self, data):
        """
        Creates a `TeamMember` object from the data sent by Discord.
        
        Parameters
        ----------
        data : `None` or `dict` of (`str`, `Any`) items
            Team member data received from Discord.
        """
        permissions = data['permissions']
        permissions.sort()
        self.permissions = permissions
        self.user = User(data['user'])
        self.state = TeamMembershipState.get(data['membership_state'])
    
    def __repr__(self):
        """Returns the team member's representation."""
        return (f'<{self.__class__.__name__} user={self.user.full_name} state={self.state.name} permissions='
            f'{self.permissions}>')
    
    def __hash__(self):
        """Returns the team member's hash value, what is equal to it's user's id."""
        return self.user.id
    
    def __eq__(self, other):
        """Returns whether the two team members are equal."""
        if type(self) is not type(other):
            return False
        
        if (self.user != other.user):
            return False
        
        if (self.state is not other.state):
            return False
        
        if (self.permissions != other.permissions):
            return False
        
        return True

class ApplicationSubEntity(DiscordEntity):
    """
    An un-typed entity stored inside of an ``Application``, as one of it's `.developers`, or `.publishers`.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the entity.
    name : `str`
        The name of the entity.
    """
    __slots__ = ('name', )
    
    def __init__(self, data):
        """
        Creates a new ``ApplicationSubEntity`` instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Developers or Publisher data.
        """
        self.id = int(data['id'])
        self.name = data['name']
    
    def __repr__(self):
        """Returns the entity's representation."""
        return f'<{self.__class__.__name__} {self.name!r}, id={self.id}>'
    
    def __eq__(self, other):
        """Returns whether the two entities equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.name != other.name:
            return False
        
        return True

class ApplicationExecutable:
    """
    Represents a game's executable.
    
    Attributes
    ----------
    arguments : `None` or `str`
        The arguments to start the application with. Defaults to `None`.
    is_launcher : `bool`
        Whether the application is a launcher. Defaults to `False`.
    name : `str`
        The executable's name.
    os : `str`
        The operation system, the executable is for.
    """
    __slots__ = ('arguments', 'is_launcher', 'name', 'os')
    
    def __init__(self, data):
        """
        Creates a new application executable with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Executable data.
        """
        self.name = data['name']
        self.os = data['os']
        self.arguments = data.get('arguments')
        self.is_launcher = data.get('is_launcher', False)
    
    def __repr__(self):
        """Returns the executable's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ' ',
            repr(self.name),
            ', os=',
            repr(self.os),
                ]
        
        arguments = self.arguments
        if (arguments is not None):
            result.append(', arguments=')
            result.append(repr(arguments))
        
        is_launcher = self.is_launcher
        if is_launcher:
            result.append(', is_launcher=True')
        
        result.append('>')
        
        return ''.join(result)
    
    def __eq__(self, other):
        """Returns whether the two executables are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.name != other.name:
            return False
        
        if self.os != other.os:
            return False
        
        if self.arguments != other.arguments:
            return False
        
        if self.is_launcher != other.is_launcher:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the entity's hash."""
        result = hash(self.name) ^ hash(self.os)
        
        arguments = self.arguments
        if (arguments is not None):
            result ^= hash(arguments)
        
        if self.is_launcher:
            result ^= (1<<15)
        
        return result

class ThirdPartySKU:
    """
    Represents a third party Stock Keeping Unit.
    
    distributor : `str`
        The distributor of the SKU.
    id : `str`
        The identifier of the third party SKU.
    sku : `str`
        Might be same as `.id`.
    """
    __slots__ = ('distributor', 'id', 'sku',)
    
    def __init__(self, data):
        """
        creates a new third party SKU object from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            SKU data.
        """
        self.distributor = data['distributor']
        self.id = data['id']
        self.sku = data['sku']
    
    def __repr__(self):
        """Returns the SKU's representation."""
        return f'<{self.__class__.__name__} distributor={self.distributor!r}, id={self.id!r}, sku={self.sku!r}>'
    
    def __eq__(self, other):
        """Returns whether the two SKU-s are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.distributor != other.distributor:
            return False
        
        if self.id != other.id:
            return False
        
        if self.sku != other.sku:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the sku's hash."""
        result = hash(self.distributor)
        
        id_ = self.id
        result ^= hash(id_)
        
        sku = self.sku
        if sku != id_:
            result ^= hash(sku)
        
        return result

class EULA(DiscordEntity, immortal=True):
    """
    Represents a Discord end-user license agreement
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the eula.
    content : `str`
        The eula's content.
    name : `str`
        The eula's name.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = ('id', 'content', 'name')
    def __new__(cls, data):
        """
        Creates a new eula instance from the given parameters.
        
        If the eula already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Eula data.
        """
        eula_id = int(data['id'])
        
        try:
            self = EULAS[eula_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = eula_id
            self._update_no_return(data)
            
            EULAS[eula_id] = self
        
        return self
    
    def _update_no_return(self, data):
        """
        Updates the eula with the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        self.content = data['content']
        self.name = data['name']
    
    def __repr__(self):
        """Returns the eula's representation"""
        return f'<{self.__class__.__name__} {self.name!r}, id={self.id}>'



# -*- coding: utf-8 -*-
__all__ = ('AuditLog', 'AuditLogEntry', 'AuditLogIterator', 'AuditLogChange', )

from ..env import API_VERSION

from .utils import Unknown, now_as_id, id_to_time
from .client_core import CHANNELS, USERS, ROLES, MESSAGES
from .permission import Permission
from .color import Color
from .user import User
from .webhook import Webhook
from .role import PermissionOverwrite
from .integration import Integration
from .guild import SystemChannelFlag, Guild
from .bases import Icon
from .preinstanced import AuditLogEvent, VerificationLevel, ContentFilterLevel, MessageNotificationLevel, VoiceRegion, \
    MFA
from.client_utils import maybe_snowflake

Client = NotImplemented

class AuditLog:
    """
    Whenever an admin action is performed on the API, an audit log entry is added to the respective guild's audit
    logs. This class represents a requested  collections of these entries.
    
    Attributes
    ----------
    guild : ``Guild``
        The audit logs' respective guild.
    entries : `list` of ``AuditLogEntry``
        A list of audit log entries, what the audit log contains.
    users : `dict` of (`int`, (``Client`` or ``User``)) items
        A dictionary, what contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary what contains the mentioned webhook by the audit log's entries. The keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary what contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    """
    __slots__ = ('guild', 'entries', 'users', 'webhooks', 'integrations', )
    def __init__(self, data, guild):
        """
        Creates an ``AuditLog`` instance from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        guild : ``Guild``
            The respective guild of the audit logs.
        """
        self.guild = guild
        
        self.users = users = {}
        try:
            user_datas = data['users']
        except KeyError:
            pass
        else:
            for user_data in user_datas:
                user = User(user_data)
                users[user.id] = user
        
        self.webhooks = webhooks = {}
        try:
            webhook_datas = data['webhook']
        except KeyError:
            pass
        else:
            for webhook_data in webhook_datas:
                webhook = Webhook(webhook_data)
                webhooks[webhook.id] = webhook
        
        self.integrations = integrations = {}
        try:
            integration_datas = data['integrations']
        except KeyError:
            pass
        else:
            for integration_data in integration_datas:
                integration = Integration(integration_data)
                integrations[integration.id] = integration
        
        self.entries = entries = []
        try:
            entry_datas = data['audit_log_entries']
        except KeyError:
            pass
        else:
            for entry_data in entry_datas:
                entries.append(AuditLogEntry(entry_data, self))
    
    def __iter__(self):
        """Iterates over the audit log's entries."""
        return iter(self.entries)
    
    def __reversed__(self):
        """Reversed iterator over the audit log's entries."""
        return reversed(self.entries)
    
    def __len__(self):
        """Returns the amount of entries, what the audit lgo contain."""
        return len(self.entries)
    
    def __getitem__(self,index):
        """Returns the specific audit log entry at the given index."""
        return self.entries.__getitem__(index)
    
    def __repr__(self):
        """Returns the representation of the Audit log."""
        return f'<{self.__class__.__name__} of {self.guild.name}, length={len(self.entries)}>'

class AuditLogIterator:
    """
    An async iterator over a guild's audit logs.
    
    Attributes
    ----------
    guild : ``Guild``
        The audit log iterator's respective guild.
    entries : `list` of ``AuditLogEntry``
        A list of the already received audit log entries.
    users : `dict` of (`int`, (``Client`` or ``User``)) items
        A dictionary, what contains the mentioned users by the audit log's entries. The keys are the `id`-s of the
        users, meanwhile the values are the users themselves.
    webhooks : `dict` of (`int`, ``Webhook``) items
        A dictionary what contains the mentioned webhook by the audit log's entries. They keys are the `id`-s of the
        webhooks, meanwhile the values are the values themselves.
    integrations : `dict` of (`int`, ``Integration``) items
        A dictionary what contains the mentioned integrations by the audit log's entries. The keys are the `id`-s of
        the integrations, meanwhile the values are the integrations themselves.
    _data : `dict` of (`str`, `Any`) items
        Data to be sent to Discord when requesting an another audit log chunk. Contains some information, which are not
        stored by any attributes of the audit log iterator, these are the filtering `user` and `event` options.
    _index : `int`
        The next audit log entries index to yield.
    client : ``Client``
        The client, who will execute the api requests.
    """
    __slots__ = ('guild',  'entries', 'users', 'webhooks', 'integrations', '_data', '_index', 'client', )
    
    async def __new__(cls, client, guild, user=None, event=None):
        """
        Creates an audit log iterator with the given arguments.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will execute the api requests.
        guild : ``Guild`` or `int` instance
            The guild, what's audit logs will be requested.
        user : `None`, ``Client``, ``User`` or `int` instance, Optional
            Whether the audit logs should be filtered only to those, which were created by the given user.
        event : `None`, ``AuditLogEvent`` or `int` instance, Optional
            Whether the audit logs should be filtered only on the given event.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, nor as `int` instance.
            - If `user` was not given neither as `None`, ``User``, ``Client`` nor as `int` instance.
            - If `event` as not not given neither as `None`, ``AuditLogEvent`` nor as `int` instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = {
            'limit' : 100,
            'before': now_as_id(),
                }
        
        if isinstance(guild, Guild):
            guild_id = guild.id
        else:
            guild_id = maybe_snowflake(guild)
            if guild_id is None:
                raise TypeError(f'`guild_or_discovery` can be `{Guild.__name__}` or `int` instance, got '
                    f'{guild.__class__.__name__}.')
            
            guild = None
        
        if (user is not None):
            if isinstance(user, (User, Client)):
                user_id = user.id
            
            else:
                user_id = maybe_snowflake(user)
                if user_id is None:
                    raise TypeError(f'`user` can be given as `{User.__name__}`, `{Client.__name__}` or `int` instance, '
                        f'got {user.__class__.__name__}.')
            
            data['user_id'] = user_id
        
        if (event is not None):
            if isinstance(event, AuditLogEvent):
                event_value = event.value
            elif isinstance(event, int):
                event_value = event
            else:
                raise TypeError(f'`event` can be given as `None`, `{AuditLogEvent.__name__}` or `int` instance, got '
                    f'{event.__class__.__name__}.')
            
            data['action_type'] = event_value
        
        if guild is None:
            log_data = await client.http.audit_log_get_chunk(guild_id, data)
            if guild is None:
                guild = Guild.precreate(guild_id)
        else:
            log_data = None
        
        self = object.__new__(cls)
        self._data = data
        self._index = 0
        self.client = client
        self.guild = guild
        self.entries = []
        self.users = {}
        self.webhooks = {}
        self.integrations = {}
        
        if (log_data is not None):
            self._process_data(log_data)
        
        return self
    
    async def load_all(self):
        """
        Loads all not yet loaded audit logs of the audit log iterator's guild.
        
        This method is a coroutine.
        """
        entries = self.entries
        client = self.client
        http = client.http
        data = self._data
        
        while True:
            if entries:
                data['before'] = entries[-1].id
            
            log_data = await http.audit_log_get_chunk(self.guild.id, data)
            
            try:
                self._process_data(log_data)
            except StopAsyncIteration:
                return
            
            if len(entries)%100:
                return
    
    def transform(self):
        """
        Converts the audit log iterator to an audit log object.
        
        Returns
        -------
        audit_log : ``AuditLog``
        """
        result = object.__new__(AuditLog)
        result.guild = self.guild
        result.entries = self.entries
        result.users = self.users
        result.webhooks = self.webhooks
        result.integrations = self.integrations
        return result
    
    def __aiter__(self):
        """Returns self and resets the `.index`."""
        self._index = 0
        return self
    
    async def __anext__(self):
        """
        Yields the next entry of the audit log iterator.
        
        This method is a coroutine.
        """
        ln = len(self.entries)
        index = self._index
        
        if index < ln:
            self._index += 1
            return self.entries[index]
        
        if index%100:
            raise StopAsyncIteration
        
        data = self._data
        if ln:
            data['before'] = self.entries[ln-1].id
        
        log_data = await self.client.http.audit_log_get_chunk(self.guild.id, data)
        self._process_data(log_data)
        self._index +=1
        return self.entries[index]

    def __repr__(self):
        """Returns the representation of the audit log iterator."""
        return f'<{self.__class__.__name__} of {self.client.full_name} at {self.guild.name}>'
    
    def _process_data(self, data):
        """
        Processes a batch of audit log data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        try:
            entry_datas = data['audit_log_entries']
        except KeyError:
            raise StopAsyncIteration from None
        
        if not entry_datas:
            raise StopAsyncIteration
        
        users = self.users
        try:
            users_data = data['users']
        except KeyError:
            pass
        else:
            for user_data in users_data:
                user = User(user_data)
                users[user.id] = user
        
        webhooks = self.webhooks
        try:
            webhooks_data = data['webhook']
        except KeyError:
            pass
        else:
            for webhook_data in webhooks_data:
                webhook = Webhook(webhook_data)
                webhooks[webhook.id] = webhook
        
        integrations = self.integrations
        try:
            integration_datas = data['integrations']
        except KeyError:
            pass
        else:
            for integration_data in integration_datas:
                integration = Integration(integration_data)
                integrations[integration.id] = integration
        
        entries=self.entries
        for entry_data in entry_datas:
            entries.append(AuditLogEntry(entry_data, self))


def convert_detail_days(key, value, all_):
    return 'days', int(value)

def convert_detail_users_removed(key, value, all_):
    return 'users_removed', int(value)

def convert_detail_channel(key, value, all_):
    channel_id = int(value)
    try:
        channel = CHANNELS[channel_id]
    except KeyError:
        channel = Unknown('Channel',channel_id)
    return 'channel', channel

def convert_detail_message(key, value, all_):
    message_id = int(value)
    try:
        message = MESSAGES[message_id]
    except KeyError:
        message = Unknown('Message',message_id)
    
    return 'message', message

def convert_detail_amount(key, value, all_):
    return 'amount', int(value)

def convert_detail_perm_ow_target(key, value, all_):
    id_=int(value)
    try:
        type_name = all_['type']
    except KeyError:
        # broken data
        return None
    
    if type_name == 'member':
        try:
            target = USERS[id_]
        except KeyError:
            target = Unknown('User', id_)
    
    elif type_name == 'role':
        try:
            target = ROLES[id_]
        except KeyError:
            target = Unknown('Role', id_, all_.get('name', ''))
    else:
        # perm_ow type can be only member and role, so if it is else,
        # the data is broken again
        return None

def convert_detail_perm_ow_processed(key, value, all_):
    return None

DETAIL_CONVERSIONS = {
    'delete_member_days': convert_detail_days,
    'members_removed'   : convert_detail_users_removed,
    'channel_id'        : convert_detail_channel,
    'message_id'        : convert_detail_message,
    'count'             : convert_detail_amount,
    'id'                : convert_detail_perm_ow_target,
    'type'              : convert_detail_perm_ow_processed,
    'role_name'         : convert_detail_perm_ow_processed,
        }

del convert_detail_days
del convert_detail_users_removed
del convert_detail_channel
del convert_detail_message
del convert_detail_amount
del convert_detail_perm_ow_target
del convert_detail_perm_ow_processed

def DETAIL_CONVERTER_DEFAULT(key, value, all_):
    return key, value


def convert_guild(entry, parent, target_id):
    return parent.guild

def convert_channel(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.channels[target_id]
        except KeyError:
            target = Unknown('Channel', target_id)
    
    return target

def convert_user(entry, parent, target_id):
    # target_id can be None for any reason
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.users[target_id]
        except KeyError:
            target = Unknown('User', target_id)
    
    return target

def convert_role(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target = int(target_id)
        try:
            target = parent.guild.roles[target_id]
        except KeyError:
            target = Unknown('Role', target_id)
    
    return target

def convert_invite(entry, parent, target_id):
    #every other data is at #change
    for change in entry.changes:
        if change.attr!='code':
            continue
        
        if entry.type is AuditLogEvent.INVITE_DELETE:
            code = change.before
        else:
            code = change.after
        break
    
    else:
        code = '' #malformed ?
    
    return Unknown('Invite', code)
    
def convert_webhook(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.webhooks[target_id]
        except KeyError:
            target = Unknown('Webhook', target_id)
    
    return target

def convert_emoji(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.guild.emojis[target_id]
        except KeyError:
            target = Unknown('Emoji', target_id)
    
    return target

def convert_message(entry, parent, target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = MESSAGES[target_id]
        except KeyError:
            target = Unknown('Message', target_id)
    
    return target

def convert_integration(entry, parent,target_id):
    if target_id is None:
        target = None
    else:
        target_id = int(target_id)
        try:
            target = parent.integrations[target_id]
        except KeyError:
            target = Unknown('Integration', target_id)
    
    return target

CONVERSIONS = (
    convert_guild,
    convert_channel,
    convert_user,
    convert_role,
    convert_invite,
    convert_webhook,
    convert_emoji,
    convert_message,
    convert_integration,
        )

del convert_guild
del convert_channel
del convert_user
del convert_role
del convert_invite
del convert_webhook
del convert_emoji
del convert_message
del convert_integration


class AuditLogEntry:
    """
    Represents an entry of an ``AuditLog``.
    
    Attributes
    ----------
    changes : `list` of ``AuditLogChange``
        The changes of the entry.
    details : `None` or `dict` of (`str`, `Any`) items
        Additional information for a specific action types.
    id : `int`
        The unique identifier number of the entry.
    reason : `None` or `str`
        The reason provided with the logged action.
    target : `None`, ``Guild``,  ``ChannelGuildBase`` instance, ``User``, ``Client``, ``Role``, ``Webhook``,
            ``Emoji``, ``Message``, ``Integration``, ``Unknown``
        The target entity of the logged action. If the entity is not found, it will be set as an ``Unknown`` instance.
        It can also happen, that target entity is not provided, then target will be set as `None`.
    type : ``AuditLogEvent``
        The event type of the logged action.
    user : `None`, ``Client``, ``User``
        The user, who executed the logged action. If no user is provided then can be `None` as well.
    """
    __slots__ = ('changes', 'details', 'id', 'reason', 'target', 'type', 'user',)
    def __init__(self, data, parent):
        """
        Creates an audit log entry, from entry data sent inside of an ``AuditLog``'s data and from the audit itself.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        parent : ``AuditLog`` or ``AuditLogIterator``
            The parent of the entry, what contains the respective guild, the included users, webhooks and the
            integrations to work with.
        """
        self.id = int(data['id'])
        self.type = AuditLogEvent.get(int(data['action_type']))
        
        options = data.get('options')
        if (options is None):
            details = None
        else:
            details = {}
            for key, value in options.items():
                result = DETAIL_CONVERSIONS.get(key, DETAIL_CONVERTER_DEFAULT)(key, value, options)
                if result is None:
                    continue
                
                key, value = result
                details[key] = value
            
            if not details:
                details = None
        
        self.details = details
        
        user_id = data.get('user_id')
        if user_id is None:
            user = None
        else:
            user = parent.users.get(int(user_id))
        self.user = user
        
        self.reason = data.get('reason')
        
        change_datas = data.get('changes')
        if (change_datas is None):
            changes = None
        else:
            changes = []
            for change_data in change_datas:
                try:
                    key = change_data['key']
                except KeyError: #malformed?
                    continue
                
                change = TRANSFORMERS.get(key, transform_nothing)(key, change_data)
                if (change is not None):
                    changes.append(change)
            
            if not changes:
                changes = None
        
        self.changes = changes
        
        self.target = CONVERSIONS[self.type.value//10](self, parent, data.get('target_id', None))
    
    @property
    def created_at(self):
        """
        When the audit log entry was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_time(self.id)
    
    def __repr__(self):
        """Returns the representation of the audit log entry."""
        result = [
            '<',self.__class__.__name__,
            ' id=',repr(self.id),
            ', type=',self.type.name,
                ]
        
        result.append(', user=')
        user = self.user
        if user is None:
            user_repr = 'None'
        else:
            user_repr = user.full_name
        result.append(user_repr)
        
        result.append(', target=')
        result.append(repr(self.target))
        
        result.append(', change count=')
        changes = self.changes
        if changes is None:
            change_amount_repr = '0'
        else:
            change_amount_repr = repr(len(self.changes))
        result.append(change_amount_repr)
        
        reason = self.reason
        if reason is not None:
            result.append(', reason=')
            # use repr to escape special inserted characters
            result.append(repr(reason))
        
        details = self.details
        if details is not None:
            result.append(', details=')
            result.append(repr(details))
        
        result.append('>')
        
        return ''.join(result)


def transform_nothing(name, data):
    change = AuditLogChange()
    change.attr = name
    change.before = data.get('old_value')
    change.after = data.get('new_value')
    return change

def transform_deprecated(name, data):
    return None

def transform_icon(name, data):
    change = AuditLogChange()
    if name == 'splash_hash':
        name = 'invite_splash'
    else:
        name = name[:-5]
    change.attr = name
    
    change.before = Icon.from_base16_hash(data.get('old_value'))
    
    change.after = Icon.from_base16_hash(data.get('new_value'))
    
    return change

def transform_bool__separated(name, data):
    change = AuditLogChange()
    change.attr = 'separated'
    change.before = data.get('old_value')
    change.after = data.get('new_value')
    return change

def transform_channel(name, data):
    change = AuditLogChange()
    change.attr = name[:-3]
    
    value = data.get('old_value')
    if value is None:
        change.before=None
    else:
        value = int(value)
        try:
            before = CHANNELS[value]
        except KeyError:
            before = Unknown('Channel', value)
        change.before = before
    
    value = data.get('new_value')
    if value is None:
        change.after = None
    else:
        value = int(value)
        try:
            after = CHANNELS[value]
        except KeyError:
            after = Unknown('Channel', value)
        change.after = after
    
    return change

def transform_color(name, data):
    change = AuditLogChange()
    change.attr = 'color'
    value = data.get('old_value')
    change.before = None if value is None else Color(value)
    value = data.get('new_value')
    change.after = None if value is None else Color(value)
    return change

def transform_content_filter(name, data):
    change = AuditLogChange()
    change.attr = 'content_filter'
    value = data.get('old_value')
    change.before = None if value is None else ContentFilterLevel.get(value)
    value = data.get('new_value')
    change.after = None if value is None else ContentFilterLevel.get(value)
    return change

def transform_int__days(name, data):
    change = AuditLogChange()
    change.attr = 'days'
    change.before = data.get('old_value')
    change.after = data.get('new_value')
    return change

def transform_int__slowmode(name, data):
    change = AuditLogChange()
    change.attr = 'slowmode'
    change.before = data.get('old_value')
    change.after = data.get('new_value')
    return change

def transform_message_notification(name, data):
    change  =AuditLogChange()
    change.attr = 'message_notification'
    before = data.get('old_value')
    change.before = None if before is None else MessageNotificationLevel.get(before)
    after = data.get('new_value')
    change.after = None if before is None else MessageNotificationLevel.get(after)
    return change

def transform_mfa(name, data):
    change = AuditLogChange()
    change.attr = 'mfa'
    before = data.get('old_value')
    change.before = None if before is None else MFA.get(before)
    after = data.get('new_value')
    change.after = None if before is None else MFA.get(after)
    return change

def transform_overwrites(name, data):
    change = AuditLogChange()
    change.attr = 'overwrites'
    value = data.get('old_value')
    change.before = None if value is None else [PermissionOverwrite(ow_data) for ow_data in value]
    value = data.get('new_value')
    change.after = None if value is None else [PermissionOverwrite(ow_data) for ow_data in value]
    return change

def transform_permission(name, data):
    change = AuditLogChange()
    
    if name.endswith('_new'):
        name = name[:-4]
    
    change.attr = name
    value = data.get('old_value')
    change.before = None if value is None else Permission(value)
    value = data.get('new_value')
    change.after = None if value is None else Permission(value)
    return change

def transform_region(name, data):
    change = AuditLogChange()
    change.attr = 'region'
    before = data.get('old_value')
    change.before = None if before is None else VoiceRegion.get(before)
    after = data.get('new_value')
    change.after = None if before is None else VoiceRegion.get(after)
    return change

def transform_role(name, data):
    change = AuditLogChange()
    change.attr = 'role'
    roles = []
    if name == '$add':
        before = None
        after = roles
    else:
        before = roles
        after = None
    
    change.before = before
    change.after = after
    
    for element in data['new_value']:
        role_id = int(element['id'])
        try:
            role = ROLES[role_id]
        except KeyError:
            role = Unknown('Role', role_id, element['name'])
        roles.append(role)

    return change

def transform_snowflake(name, data):
    change = AuditLogChange()
    change.attr = name
    value = data.get('old_value')
    change.before = None if value is None else int(value)
    value = data.get('new_value')
    change.after = None if value is None else int(value)
    return change

def transform_str__vanity_code(name, data):
    change = AuditLogChange()
    change.attr = 'vanity_code'
    change.before = data.get('old_value')
    change.after = data.get('new_value')
    return change

def transform_system_channel_flags(name, data):
    change = AuditLogChange()
    change.attr = 'system_channel_flags'
    before = data.get('old_value')
    change.before = None if before is None else SystemChannelFlag(before)
    after = data.get('new_value')
    change.after = None if before is None else SystemChannelFlag(after)
    return change

def transform_type(name, data):
    # If we talk about permission overwrite, type can be `str` too, what we ignore
    before = data.get('old_value')
    if type(before) is str:
        return
    
    after = data.get('new_value')
    if type(after) is str:
        return
    
    change = AuditLogChange()
    change.attr = 'type'
    change.before = before
    change.after = after
    return change

def transform_user(name, data):
    change = AuditLogChange()
    change.attr = name[:-3]
    value = data.get('old_value')
    if value is None:
        before = None
    else:
        value = int(value)
        try:
            before = USERS[value]
        except KeyError:
            before = Unknown('User', value)
    change.before = before
    
    value = data.get('new_value')
    if value is None:
        after = None
    else:
        value = int(value)
        try:
            after = USERS[value]
        except KeyError:
            after = Unknown('User', value)
    change.after = after
    
    return change

def transform_verification_level(name, data):
    change = AuditLogChange()
    change.attr = 'verification_level'
    value = data.get('old_value')
    change.before = None if value is None else VerificationLevel.value[value]
    value = data.get('new_value')
    change.after = None if value is None else VerificationLevel.value[value]
    return change

TRANSFORMERS = {
    '$add'                  : transform_role,
    '$remove'               : transform_role,
    'account_id'            : transform_snowflake,
    'afk_channel_id'        : transform_channel,
    'allow'                 : transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'allow_new'             : transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'application_id'        : transform_snowflake,
    'avatar_hash'           : transform_icon,
    'banner_hash'           : transform_icon,
    # bitrate (int)
    'channel_id'            : transform_channel,
    # code (str)
    'color'                 : transform_color,
    # deaf (bool)
    # description (None or str)
    'default_message_notifications':transform_message_notification,
    'deny'                  : transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'deny_new'              : transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'discovery_splash_hash' : transform_icon,
    # enable_emoticons (bool)
    # expire_behavior (int)
    # expire_grace_period (int)
    'explicit_content_filter':transform_content_filter,
    'hoist'                 : transform_bool__separated,
    'icon_hash'             : transform_icon,
    'id'                    : transform_snowflake,
    'inviter_id'            : transform_user,
    # mentionable (bool)
    # max_age (int)
    # max_uses (int)
    'mfa_level'             : transform_mfa,
    # mute (mute)
    # name (str)
    # nick (None or str)
    # nsfw (bool)
    'owner_id'              : transform_user,
    # position (int)
    'prune_delete_days'     : transform_int__days,
    'permission_overwrites' : transform_overwrites,
    'permissions'           : transform_deprecated if API_VERSION in (6, 7) else transform_permission,
    'permissions_new'       : transform_permission if API_VERSION in (6, 7) else transform_deprecated,
    'public_updates_channel_id' : transform_channel,
    'rate_limit_per_user'   : transform_int__slowmode,
    'region'                : transform_region,
    'rules_channel_id'      : transform_channel,
    'splash_hash'           : transform_icon,
    'system_channel_id'     : transform_channel,
    'system_channel_flags'  : transform_system_channel_flags,
    # temporary (bool)
    # topic (str)
    'type'                  : transform_type,
    # uses (int)
    'vanity_url_code'       : transform_str__vanity_code,
    'verification_level'    : transform_verification_level,
    'widget_channel_id'     : transform_channel,
    # widget_enabled (bool)
        }

del transform_deprecated
del transform_icon
del transform_bool__separated
del transform_channel
del transform_color
del transform_content_filter
del transform_int__days
del transform_int__slowmode
del transform_message_notification
del transform_mfa
del transform_overwrites
del transform_permission
del transform_region
del transform_role
del transform_snowflake
del transform_str__vanity_code
del transform_system_channel_flags
del transform_type
del transform_user
del transform_verification_level


class AuditLogChange:
    """
    A change of an ``AuditLogEntry``.
    
    Attributes
    ----------
    attr : `str`
        The name of the attribute, what changed of the target entity.
    before : `Any`
        The changed attribute's original value. Defaults to `None`.
    after : `Any`
        The changed attribute's new value. Defaults to `None`.
    
    Notes
    -----
    The value of `before` and `after` depending on the value of `attr`. These are:
    
    +---------------------------+-------------------------------------------+
    | attr                      | before / after                            |
    +===========================+===========================================+
    | account_id                | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | afk_channel               | `None` or ``ChannelVoice``                |
    +---------------------------+-------------------------------------------+
    | allow                     | `None` or ``Permission``                  |
    +---------------------------+-------------------------------------------+
    | application_id            | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | avatar                    | `None` or ``Icon``                        |
    +---------------------------+-------------------------------------------+
    | banner                    | `None` or ``Icon``                        |
    +---------------------------+-------------------------------------------+
    | bitrate                   | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | channel                   | `None` or ``ChannelGuildBase`` instance   |
    +---------------------------+-------------------------------------------+
    | code                      | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | color                     | `None` or ``Color``                       |
    +---------------------------+-------------------------------------------+
    | content_filter            | `None` or ``ContentFilterLevel``          |
    +---------------------------+-------------------------------------------+
    | days                      | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | deaf                      | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | description               | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | deny                      | `None` or ``Permission``                  |
    +---------------------------+-------------------------------------------+
    | discovery_splash          | `None` or ``Icon``                        |
    +---------------------------+-------------------------------------------+
    | enable_emoticons          | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | expire_behavior           | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | expire_grace_period       | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | icon                      | `None` or ``Icon``                        |
    +---------------------------+-------------------------------------------+
    | id                        | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | invite_splash             | `None` or ``Icon``                        |
    +---------------------------+-------------------------------------------+
    | inviter                   | `None`, ``User`` or ``Client``            |
    +---------------------------+-------------------------------------------+
    | mentionable               | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | max_age                   | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | max_uses                  | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | message_notification      | `None` or ``MessageNotificationLevel``    |
    +---------------------------+-------------------------------------------+
    | mfa                       | `None` or ``MFA``                         |
    +---------------------------+-------------------------------------------+
    | mute                      | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | name                      | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | nick                      | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | nsfw                      | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | owner                     | `None`, ``User`` or ``Client``            |
    +---------------------------+-------------------------------------------+
    | position                  | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | public_updates_channel    | `None` or ``ChannelText``                 |
    +---------------------------+-------------------------------------------+
    | overwrites                | `None` or `list` of ``PermissionOverwrite``            |
    +---------------------------+-------------------------------------------+
    | permissions               | `None` or ``Permission``                  |
    +---------------------------+-------------------------------------------+
    | region                    | `None` or ``VoiceRegion``                 |
    +---------------------------+-------------------------------------------+
    | role                      | `None` or `list` of ``Role``              |
    +---------------------------+-------------------------------------------+
    | rules_channel             | `None` or ``ChannelText``                 |
    +---------------------------+-------------------------------------------+
    | separated                 | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | slowmode                  | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | system_channel            | `None` or ``ChannelText``                 |
    +---------------------------+-------------------------------------------+
    | system_channel_flags      | `None` or ``SystemChannelFlag``           |
    +---------------------------+-------------------------------------------+
    | temporary                 | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    | topic                     | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | type                      | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | uses                      | `None` or `int`                           |
    +---------------------------+-------------------------------------------+
    | vanity_code               | `None` or `str`                           |
    +---------------------------+-------------------------------------------+
    | verification_level        | `None` or ``VerificationLevel``           |
    +---------------------------+-------------------------------------------+
    | widget_channel            | `None` or ``ChannelText``                 |
    +---------------------------+-------------------------------------------+
    | widget_enabled            | `None` or `bool`                          |
    +---------------------------+-------------------------------------------+
    """
    __slots__ = ('attr', 'before', 'after', )
    
    def __repr__(self):
        """Returns the representation of the audit log change."""
        return f'{self.__class__.__name__}(attr={self.attr!r}, before={self.before!r}, after={self.after!r})'
